/**
 * 
 * Copyright (C) 2001 Alexandre. N. Safiullin <alex_@unis-ru.com> 
 * 
 * The contents of this file are subject to the GNU General Public License
 * Version 2 (the License). You may not copy or use this
 * file, in either source code or executable form, except in compliance
 * with the License.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

//     In addition, as a special exception, Alexandre N. Safiullin 
//     gives permission to link the code of this program with
//     the JECL (Jabber External Component Library, jabber.tigris.org)
//     (or with modified versions of JECL that use the
//     same license as JECL), and distribute linked combinations including
//     the two.  You must obey the GNU General Public License in all
//     respects for all of the code used other than JECL.  If you modify
//     this file, you may extend this exception to your version of the
//     file, but you are not obligated to do so.  If you do not wish to
//     do so, delete this exception statement from your version.

/* File: IcqTrans.cpp -> a part of the "icqv7-t" project 
   Created: Thu Nov 15 20:39:29 2001
		by Alexandre N. Safiullin <alex_@unis-ru.com>
   Contains: implementation of IcqTrans */

#include <IcqTrans.h>
#include <it_Client.h>
#include <icqv7t.h>

using namespace std;

bedrock::debug::Tracer t("icqv7ext");

// Constructor
IcqTrans::IcqTrans(const std::string& serviceid, 
			       const std::string& password, 
			       const std::string& hostname, 
			       unsigned int port, bool outgoing_dir) :
	_id(serviceid), _tpool(5), timer(1), _threadkey(-1), _watcher(_tpool,10), selecter(_tpool),
	router(_watcher, *this, outgoing_dir, 0), unconfigured(true)
{
	/* now time to do standard things: */
    // Connection type (accept/connect)
	t<<"ID: "<<_id<<endl
	 <<"Host: "<<hostname<<endl
	 <<"Port: "<<port<<endl
	 <<"Password: "<<password<<endl;

	Address router_addr(hostname, "_jabber._tcp", port);
	
	router.connect(_id, password, router_addr);
}

IcqTrans::~IcqTrans(){
}

void IcqTrans::deliver(MyElement* me, bool mangle){
	// let's make -j SNAT ;)
    if(mangle)xmlnode_put_attrib(me->jp->x,"from",internal_part.c_str());

	router.deliver(me);
}

void IcqTrans::deliver(Packet* p){
	RouterConnection<Packet,Packet>* fuckingCPP=&router;
	fuckingCPP->deliver(p);
}	

// Router event callbacks
void IcqTrans::onRouterConnected()
{
    t << "Router is now connected." << endl;
	
	/* requesting configuration */
	string ns("jabber:config:");
    ns += _id;
    judo::Element *request = new judo::Element("xdb");
    request->putAttrib("type", "get");
    request->putAttrib("to", "config@-internal");
    request->putAttrib("from", _id);
    request->putAttrib("ns", ns.c_str());
    deliver(new Packet(request));

}

void IcqTrans::onRouterDisconnected()
{
    cerr << "Router is now disconnected." << endl;
    bedrock::Application::stop(1, "ICQv7-t: Router connection lost");
}

void IcqTrans::onRouterError()
{
    cerr << "Router error occurred." << endl;
    router.disconnect();
}

void IcqTrans::onRouterPacket(jax::Packet* packet)
{
    t << "Received packet from router: " << packet->toString() << endl;

	if(unconfigured) { /* first packet is configuration arrived */
		configure(packet);
		return;
	}

 	MyElement *myelem;
	try {
		myelem=new MyElement(packet);

	} catch(const MyElement::exception::BadElement& e){
		t<<"Internal error: Packet is invalid!"<<endl;
		delete myelem; //ignore invalid packets
		return;
	}
	
	jpacket jp=myelem->jp;

	if(myelem->ptyp==p_ROUTE){
		receivedRoute(packet);
		return;
	}
	
	if (!jp->from || !jp->from->user || 
		jp->type == JPACKET_UNKNOWN || myelem->subtype == JPACKET__ERROR) {
		/* ignore invalid packets */
		t<<"Internal error: Packet is invalid!"<<endl;
		delete myelem;
		return;
    }

#ifdef _ADMIN_IFACE
	if(admin==jid_full(jid_user(jp->from))&&checkAdmin(myelem))return;
#endif

	SessionHandle p=sessions.find(jid_full(jid_user(jp->from)));
    if (p!=sessions.end()) {
        if (p->second->exit_flag) {
            if (jp->type != JPACKET_PRESENCE) {
				t<<"Internal error, we shouldn't received it :( Let's ignore"<<endl;
				delete myelem;
            }
            else
				delete myelem;
        }
        else p->second->packetReceived(myelem);
	}
    else
        unknown(myelem);
	t<<"----------------exit from routerPacket----"<<endl;
}

void IcqTrans::sendSmsContacts(const string& id) {
	SessionHandle p=sessions.find(id);
	queue<MyElement* > nonhandled=_sms_contacts[id];
	
	t<<"sending sms contacts to: "<<id<<" session"<<endl;

	while(!nonhandled.empty()){
		t<<"sending sms contact"<<endl;
		p->second->packetReceived(nonhandled.front());
		nonhandled.pop();
	}
	
	_sms_contacts.erase(id);
}

void IcqTrans::unknown(MyElement* myelem) {
	if(_sms_id==string(myelem->jp->to->server)){
		if(myelem->jp->type==JPACKET_PRESENCE &&
		   myelem->subtype==JPACKET__PROBE)
			_sms_contacts[jid_full(jid_user(myelem->jp->from))].push(myelem);
		else
			unknown_bounce(myelem);
		return;
	}

	switch(myelem->jp->type){
		
	case JPACKET_IQ:
		unknown_iq(myelem);
		break;

	case JPACKET_MESSAGE:
	case JPACKET_S10N:
		unknown_bounce(myelem);
		break;

	case JPACKET_PRESENCE:
		if (myelem->subtype == JPACKET__AVAILABLE && myelem->jp->to->user == NULL
			/* FIXME, config issues */
			/* 	&& ti->server_ip */) {
			cerr<<"Ignoring presence packet w/o registration"<<endl<<flush;
			delete myelem;
			break;
		}
	default:
		delete myelem;
	}
}

void IcqTrans::unknown_iq(MyElement* myelem) {
	char *ns;
	jpacket jp=myelem->jp;
	
	if (myelem->jp->to->user != NULL) {
		unknown_bounce(myelem);
		return;
	}
	
	ns = xmlnode_get_attrib(jp->iq,"xmlns");
	/* FIXME: maybe Judo:: stuff should be used here */
	
	switch (myelem->subtype) {
	case JPACKET__SET:
		if (j_strcmp(ns,NS_REGISTER) == 0)
			unknown_reg_set(myelem);
		else {
			t<<"Internal error, we shouldn't received it :( Let's ignore"<<endl;
			delete myelem;
			
		}
		break;
		
	default:
		t<<"Internal error, we shouldn't received it :( Let's ignore"<<endl;
		delete myelem;
	}
}

// This is oblosolete now, 'cause we receive registration packet each time
// session is establishing.

// void IcqTrans::unknown_presence(MyElement* myelem) {
//   jpacket jp = myelem->jp;
//   xmlnode reg;
//   it_Client* s;
  
//   reg = xdb_get(ti->xc,it_xdb_id(jp->p,jp->from,jp->to->server),NS_REGISTER);
//   if (reg == NULL)
//     {
// 	  log_debug(ZONE,"Registration not found for %s",jid_full(jp->from));
// 	  xmlnode_free(jp->x);
// 	  return;
//     }
  
//     /* it's possible that the session was created while we waited for the xdb result.
//        sending two available presence fast enough can cause this */
//     s = (session) xhash_get(ti->sessions,jid_full(jid_user(jp->from)));
//     if (s != NULL)
//     {
//         log_debug(ZONE,"Session %s already created",jid_full(jp->from));
//         jp->aux1 = (void *) s;
//         mtq_send(s->q,jp->p,it_session_jpacket,(void *) jp);
//         xmlnode_free(reg);
//         return;
//     }

//     s = it_session_create(ti,jp);
//     s->type = stype_normal;
//     s->uin = it_strtouin(xmlnode_get_tag_data(reg,"username"));
//     s->passwd = it_convert_utf82windows(s->p,xmlnode_get_tag_data(reg,"password"));

//     if (s->uin == 0 || s->passwd == NULL)
//     {
//         log_warn(ti->i->id,"User %s has invalid registration setting",jid_full(jp->from));
//         xmlnode_free(jp->x);
//         return;
//     }

//     s->nick = it_convert_utf82windows(s->p,xmlnode_get_tag_data(reg,"nick"));
//     s->first = it_convert_utf82windows(s->p,xmlnode_get_tag_data(reg,"first"));
//     s->last = it_convert_utf82windows(s->p,xmlnode_get_tag_data(reg,"last"));
//     s->email = it_convert_utf82windows(s->p,xmlnode_get_tag_data(reg,"email"));

//     xmlnode_free(reg);

//     /* update users presence */
//     s->p_db = ppdb_insert(s->p_db,jp->from,jp->x);
//     s->status = it_show2status(xmlnode_get_tag_data(jp->x,"show"));

//     /* start their session */
//     mtq_send(s->q,s->p,it_session_start,(void *) s);

//     /* Echo back the users presence */
//     xmlnode_put_attrib(jp->x,"from",jid_full(jp->to));
//     xmlnode_put_attrib(jp->x,"to",jid_full(jid_user(jp->from)));
//     it_deliver(ti,jp->x);
// }

void IcqTrans::unknown_reg_set(MyElement *myelem) {

    it_Client* s;
	jpacket jp=myelem->jp;
    xmlnode q = jp->iq;
    unsigned int uin;
    char *user, *pass;

    pass = xmlnode_get_tag_data(q,"password");
    user = xmlnode_get_tag_data(q,"username");
    uin = user != NULL ? strtoul(user,NULL,10): 0;

    if (pass == NULL || (uin == 0 && user))
    {
		t<<"Bad passwd or UIN, Internal error, we shouldn't"
		 <<" received it :( Let's ignore"<<endl;
		delete myelem;

    }

	/* FIXME: convert password from UTF-8 */
    s = new it_Client(*this,jid_full(jid_user(jp->from)),uin,pass);
	
	/* let's sleep a little, maybe contact list will arrive
	   from the internal part */
	s->start();
	delete myelem;
}

void IcqTrans::confirmConnection(const string& jid){
	xmlnode tt,iq=jutil_iqnew(JPACKET__SET,NS_CONNECTION);
	tt=xmlnode_get_tag(iq,"query");

	xmlnode_insert_cdata(xmlnode_insert_tag(tt,"jid"),jid.c_str(),jid.size());
	xmlnode_put_attrib(iq,"to",internal_part.c_str());
	xmlnode_put_attrib(iq,"from",_id.c_str());


	t<<"confirmConnection for "<<jid<<": "<<endl
	 <<"packet: "<<xmlnode2str(iq)<<endl;
	
	deliver(new MyElement(iq));
}

void IcqTrans::confirmDisconnect(const string& jid, const string& message,
								 int errorcode, bool fatal){

	xmlnode tt,iq=jutil_iqnew(JPACKET__ERROR,NS_CONNECTION);
	tt=xmlnode_get_tag(iq,"query");

	xmlnode_insert_cdata(xmlnode_insert_tag(tt,"jid"),jid.c_str(),jid.size());

	if(fatal)xmlnode_insert_tag(tt,"fatal");

	xmlnode_put_attrib(iq,"to",internal_part.c_str());
	xmlnode_put_attrib(iq,"from",_id.c_str());

	xmlnode_insert_cdata(tt=xmlnode_insert_tag(iq,"error"),message.c_str(),message.size());
	xmlnode_put_attrib(tt,"code",utils::ui2string(errorcode).c_str());

	t<<"confirmDisconnect for "<<jid<<": "<<endl
	 <<"packet: "<<xmlnode2str(iq)<<endl;
	
	deliver(new MyElement(iq));
}


void IcqTrans::receivedRoute(Packet *p){
	cerr<<"Received route packet!"<<endl;
	delete p;
}

void IcqTrans::unknown_bounce(MyElement *me){
	cerr<<"unknown_bounce!"<<endl;
	delete me;
}

void IcqTrans::end(const string id){
	SessionHandle s=sessions.find(id);
	if(s==sessions.end()){
		t<<"Internal error, the session was finished by someone else!"<<endl;
		return; 
	}
	else {	
		t<<"Trying to kill the session: "<<id<<endl;
		it_Client *session=s->second;

//      it done on disconnect event
//		selecter.removeSocket(session->getSocket());

		sessions.erase(id);
		delete session;
		t<<"session deleted"<<endl;
	}
	
}


/*
void it_unknown_bounce(void *arg)
{
    jpacket jp = (jpacket) arg;
    iti ti = (iti) jp->aux1;
    xmlnode reg;

    reg = xdb_get(ti->xc,it_xdb_id(jp->p,jp->from,jp->to->server),NS_REGISTER);
    if (reg != NULL)
    {
        jutil_error(jp->x,(terror){404,"Session Not Found"});
        xmlnode_free(reg);
    }
    else
        jutil_error(jp->x,TERROR_REGISTER);

    it_deliver(ti,jp->x);
} 
*/

/* auxilary function from SCM */

static string _getConfigData(judo::Element *config, const string& tag)
{
    judo::Element *child = config->findElement(tag);
    return child ? child->getCDATA() : "";
}

/**
 * Callback from ComponentLoader. Called after configuration received.
 */


void IcqTrans::configure(Packet *p)
{
	Element *ep = p->getElement();
    if (ep->getAttrib("type") != "result")
		bedrock::Application::exit(1, "Non-configuration packet received. Exiting.");
    if(ep->empty())
		bedrock::Application::exit(1, "Configuration is empty. Exiting.");
	Element *config = dynamic_cast<judo::Element*>(*ep->begin());
	
    t << "Configuration:\n " << config->toString() << endl;
    
	internal_part=_getConfigData(config,"internal");
	if(internal_part.size()==0){
		bedrock::Application::exit(1, "the ID of my internal counterpart isn't specified in config");
	}

#ifdef _ADMIN_IFACE
	admin=_getConfigData(config,"admin");
#endif	

	_sms_id=_getConfigData(config,"sms");

	ep=config->findElement("smsshow");
	if(ep){
		_sms_show=ep->getCDATA();
		if(!_sms_show.size())_sms_show="online";
	}
	t<<"sms_show: "<<_sms_show<<endl;
	reconnect=config->findElement("reconnect")==NULL?false:true;

	charset=_getConfigData(config,"charset");
	if(charset.size()==0){
		cerr<<"You omitted 'charset' specification in your config, so send all complaints"
			<<" about inproper i18n and unreadable messages to /dev/null"<<endl
			<<"Defaulting charset to 'windows-1252' (_*Latin1*_)"<<endl;
		charset="windows-1252";
	}

	
	iconv_t d1=iconv_open("UTF-8",charset.c_str());
	if(d1==(iconv_t)-1){
		cerr<<"Can't allocate iconv-descriptor for pair UTF-8 vs "<<charset<<endl
			<<"your system lacks proper iconv realization or specified charset is illegal"<<endl;
		bedrock::Application::exit(1, "Can't allocate iconv-descriptor");		
	}

	iconv_t d2=iconv_open(charset.c_str(),"UTF-8");
	if(d2==(iconv_t)-1){
		cerr<<"Can't allocate iconv-descriptor for pair "<<charset<<" vs UTF-8"<<endl
			<<"your system lacks proper iconv realization or specified charset is illegal"<<endl;
		bedrock::Application::exit(1, "Can't allocate iconv-descriptor");		
	}

	utils::converter_2UTF8=d1;
	utils::converter_unUTF8=d2;

	unconfigured=false;
	delete p;
}

#ifdef _ADMIN_IFACE

bool IcqTrans::adminCheck(MyElement* me){
	jpacket jp=me->jp;
	
	/* checks for admin packet */
	if(jp->to->user)return false;
	if(jp->type!=JPACKET_MESSAGE)return false;

	/* that's ok, it's admin packet */

	string body=xmlnode_get_tag_data(jp->x,"body");
	char* jid=new char[body.size()];
	jid[0]=0;

	if(j_strncmp(body,"??",2)==0){
		sscanf(body,"??%s",jid);
		if(jid[0]==0)adminDetailInfo(me);
		else adminDetailUser(jid,me);
	} else if(body[0]=='?'){
		sscanf(body,"?%s",jid);
		if(jid[0]==0)adminInfo(me);
		else adminUser(jid,me);
	} else {
		delete jid;
		return false;
	}

	delete jid;
	return true;
}

void IcqTrans::adminUser(char* jid, MyElement* me){

}

#endif
