/**
 * 
 * Copyright (C) 2001 Alexandre. N. Safiullin <alex_@unis-ru.com> 
 * 
 * The contents of this file are subject to the GNU General Public License
 * Version 2 (the License). You may not copy or use this
 * file, in either source code or executable form, except in compliance
 * with the License.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

//     In addition, as a special exception, Alexandre N. Safiullin 
//     gives permission to link the code of this program with
//     the JECL (Jabber External Component Library, jabber.tigris.org)
//     (or with modified versions of JECL that use the
//     same license as JECL), and distribute linked combinations including
//     the two.  You must obey the GNU General Public License in all
//     respects for all of the code used other than JECL.  If you modify
//     this file, you may extend this exception to your version of the
//     file, but you are not obligated to do so.  If you do not wish to
//     do so, delete this exception statement from your version.

/* File: IcqTrans.h -> a part of the "icqv7-t" project 
   Created: Thu Nov 15 20:31:00 2001
		by Alexandre N. Safiullin <alex_@unis-ru.com>
   Contains:  */

#ifndef IcqTrans_h_
#define IcqTrans_h_

#include <jax.h>
#include <MyElement.h>
#include <MySocketWatcher.h>

class it_Client;

using namespace jax;
using namespace bedrock::net;

class _HashString {
public:
	size_t operator()(string const &str) const { 
		return hash<char const *>()(str.c_str()); }
};

class IcqTrans :  public MyRouterConnection::EventListener {
public:
	IcqTrans(const std::string& serviceid, 
					   const std::string& password, 
					   const std::string& hostname, 
					   unsigned int port, bool outgoing_dir);
	~IcqTrans();
	

    // Router event callbacks
    void onRouterConnected();
    void onRouterDisconnected();
    void onRouterError();
    void onRouterPacket(jax::Packet* pkt);


// changes 'from'-attrib if mangle is true
	void deliver(MyElement*, bool mangle=false); 

// doesn't mangle packets
	inline void deliver(Packet* ); 
	const string& getInternalName() const { return internal_part; }
private:

	string _id;
	string _sms_id;
	string _sms_show; // show sms contacts as

	hash_map<const string, queue<MyElement*>,_HashString > _sms_contacts;
	void sendSmsContacts(const string& id);

    bedrock::ThreadPool _tpool;
	Timer timer;
	int _threadkey;

    SocketWatcher _watcher;
  /* select loop runner */
 	MySocketWatcher selecter;

  /* router */
    MyRouterConnection router;

	bool unconfigured;

  /* configurable fields */
 	std::string internal_part;
	std::string charset;
	bool reconnect;
	std::string admin;

	/* Transport-specific fields: */
	hash_map<const string,it_Client*,_HashString> sessions; //sessions by jid of its user
	typedef hash_map<const string,it_Client*,_HashString>::iterator SessionHandle;
	
	void configure(Packet* config);

	void receivedRoute(jax::Packet* packet);
	

	void confirmConnection(const string& session_owner);
	void confirmDisconnect(const string& session_owner, const string& message,
						   int errorcode, bool fatal=false);

/* The following functions deal with packet from users w/o session established */
	void unknown(MyElement* );
	void unknown_iq(MyElement* );
	void unknown_bounce(MyElement*);
	void unknown_presence(MyElement *);
	void unknown_reg_set(MyElement *);
	void unknown_reg_get(MyElement *);

/* finishes the session with given id */
	void end(const string id);
	friend class it_Client;
};

extern bedrock::debug::Tracer t;

#endif
