/**
 * 
 * Copyright (C) 2001 Alexandre. N. Safiullin <alex_@unis-ru.com> 
 * 
 * The contents of this file are subject to the GNU General Public License
 * Version 2 (the License). You may not copy or use this
 * file, in either source code or executable form, except in compliance
 * with the License.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

//     In addition, as a special exception, Alexandre N. Safiullin 
//     gives permission to link the code of this program with
//     the JECL (Jabber External Component Library, jabber.tigris.org)
//     (or with modified versions of JECL that use the
//     same license as JECL), and distribute linked combinations including
//     the two.  You must obey the GNU General Public License in all
//     respects for all of the code used other than JECL.  If you modify
//     this file, you may extend this exception to your version of the
//     file, but you are not obligated to do so.  If you do not wish to
//     do so, delete this exception statement from your version.

/* File: MyElement.cpp -> a part of the "icqv7-t" project 
   Created: Fri Nov 16 01:17:42 2001
		by Alexandre N. Safiullin <alex_@unis-ru.com>
   Contains: implementation of MyElement */


#include <MyElement.h>
#include <IcqTrans.h>

MyElement::MyElement(xmlnode x) : packet(NULL) {
	dpacket d=dpacket_new(x); // malloc'd from the pool of x (my guess)

	ptyp=d->type;
	jp=jpacket_new(x); // malloc'd from the pool of x (my guess)
	subtype=jpacket_subtype(jp);
}

MyElement::MyElement(Packet* p) : packet(p) {
	char* s=strdup(p->toString().c_str());
	xmlnode x=xmlnode_str(s,strlen(s));
	free(s);

	if(!x){
		throw exception::BadElement();
	}

	dpacket d=dpacket_new(x); // malloc'd from the pool of x (my guess)
	ptyp=d->type;
	jp=jpacket_new(x); // malloc'd from the pool of x (my guess)
	subtype=jpacket_subtype(jp);
}

MyElement::~MyElement(){
	xmlnode_free(jp->x);
	if(packet)delete packet;
}

string MyElement::toString(){
	return xmlnode2str(jp->x);
}

void MyRouterConnection::deliver(MyElement* me) {
	string s=me->toString();
	
	bedrock::MutexProxy mp(_state_lock);
	
	if (_state == Disconnected || _state == Error) {
		std::cerr << "Couldn't deliver packet: " << s << std::endl;
	}  else  {
		_socket->write(s);
	}
	delete me;
	t<<"Packet: "<<s<<endl<<" --- delivered and freed"<<endl;
}

MyRouterConnection::MyRouterConnection(bedrock::net::SocketWatcher& w,
									   EventListener& el,
									   bool outgoing,
									   unsigned int threadkey) :
	RouterConnection<Packet,Packet>(w,el,outgoing,threadkey){};

	
