/**
 * 
 * Copyright (C) 2001 Alexandre. N. Safiullin <alex_@unis-ru.com> 
 * 
 * The contents of this file are subject to the GNU General Public License
 * Version 2 (the License). You may not copy or use this
 * file, in either source code or executable form, except in compliance
 * with the License.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

//     In addition, as a special exception, Alexandre N. Safiullin 
//     gives permission to link the code of this program with
//     the JECL (Jabber External Component Library, jabber.tigris.org)
//     (or with modified versions of JECL that use the
//     same license as JECL), and distribute linked combinations including
//     the two.  You must obey the GNU General Public License in all
//     respects for all of the code used other than JECL.  If you modify
//     this file, you may extend this exception to your version of the
//     file, but you are not obligated to do so.  If you do not wish to
//     do so, delete this exception statement from your version.

/* File: MySocketWatcher.cpp -> a part of the "icqv7-t" project 
   Created: Thu Nov 15 19:57:30 2001
		by Alexandre N. Safiullin <alex_@unis-ru.com>
   Contains: implementation of MySocketWatcher */

#include <MySocketWatcher.h>
#include <it_Client.h>

/* from libicq2000 */
#include <libicq2000/events.h>

MySocketWatcher::MySocketWatcher(ThreadPool& pool, unsigned int maxsockets) :
  _started(false), _socket_cnt(maxsockets), _threadpool(pool) {
	FD_ZERO(&_read_set);
	FD_ZERO(&_write_set);
}

  
  
MySocketWatcher::~MySocketWatcher(){
	MutexProxy mp(_lock);
	if(sessions.size()==0) /* somebody forgot to removeSocket! */
		for(hash_map<int,it_Client*>::const_iterator i=sessions.begin();
			i!=sessions.end();i++)
			removeSocket(i->first);
	
	/* at this point select-thread is finished already */
}


void MySocketWatcher::addSocket(int fd, it_Client *s, int mode) {
	
	MutexProxy mp(_lock);
	if(_socket_cnt&&sessions.size()==_socket_cnt)throw exception::WatcherFull();
	
	if(sessions.find(fd)!=sessions.end()){
		t<<"BTW, such socket is already here!"<<endl;
		
		/* let's get it out */
		sessions.erase(fd);
		
		t<<"removing sockets from sets:"<<endl;
		FD_CLR(fd,&_read_set);
		t<<"removed from _read_set"<<endl;
		FD_CLR(fd,&_write_set);
		t<<"remove from _write_set"<<endl;
	}

	if(mode & SocketEvent::READ){
		t<<"Adding to read set"<<endl;
		FD_SET(fd,&_read_set);
	}
	if(mode & SocketEvent::WRITE){
		t<<"Adding to write set"<<endl;
		FD_SET(fd,&_write_set);
	}
	sessions[fd]=s;

	if(!_started){
		_started=true;


// I'm confused, it doesn't work:
//   after launching select thread main thread hangs... 
//		_threadpool.enqueue(bedrock::callback::wrap(this,&MySocketWatcher::process));

// so using old good pthread_create		
		pthread_create(&_watcher_thread, 
					   NULL,
					   MySocketWatcher::beginProcessing, 
					   this);
	}
	t<<"addSocket: addition finished"<<endl;
}

void MySocketWatcher::removeSocket(int fd) {

	MutexProxy mp(_lock);
	sessions.erase(fd);
	t<<"removing socket "<<fd<<" from sets:"<<endl;
	FD_CLR(fd,&_read_set);
	t<<"removed from _read_set"<<endl;
	FD_CLR(fd,&_write_set);
	t<<"removed from _write_set, Descriptor removed"<<endl;
}

void* MySocketWatcher::beginProcessing(void* arg){
	t<<"beginProcessing"<<endl;
	MySocketWatcher* watcher = (MySocketWatcher *)arg;
	watcher->process();
	return NULL;
}


void _my_cancellation_cleanup(void* arg)
{
    Mutex* m = (Mutex*)arg;
    m->unlock();
}

/**
 *  The main poll/process loop.
 */
void MySocketWatcher::process(void) {
	fd_set read_set, write_set;
	int c;
	timeval timeout;
	hash_map<int,it_Client*> mysessions;
	t<<"Entering select loop"<<endl<<flush;
	while (1) {
		

		/* reinitialize */ 
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;


		_lock.lock();
		read_set=_read_set;
		write_set=_write_set;
		mysessions=sessions;

		if(mysessions.size()==0){
			/* no sessions :( 
			   We will be called as soon as they appear again */

			t<<"select loop is finishing"<<endl;
			_started=false;
			_lock.unlock();
			return;
		}
		_lock.unlock();

		c=select(FD_SETSIZE,&read_set,&write_set,NULL,&timeout);

		if(c<0){
			t<<"Troubles with select: "<<strerror(errno)<<endl<<flush;
			exit(-1); 
		}
		else
			for(hash_map<int,it_Client*>::const_iterator i=mysessions.begin();
				i!=mysessions.end();i++){
				if(FD_ISSET(i->first,&read_set)){
					t<<"Calling socket_cb for read!"<<endl<<flush;
					((Client*)i->second)->socket_cb(i->first,SocketEvent::READ);
				}
				if(FD_ISSET(i->first,&write_set)){
					t<<"Calling socket_cb for write!"<<endl<<flush;
					((Client*)i->second)->socket_cb(i->first,SocketEvent::WRITE);
				}					
			}
	}
}
