/**
 * 
 * Copyright (C) 2001 Alexandre. N. Safiullin <alex_@unis-ru.com> 
 * 
 * The contents of this file are subject to the GNU General Public License
 * Version 2 (the License). You may not copy or use this
 * file, in either source code or executable form, except in compliance
 * with the License.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

//     In addition, as a special exception, Alexandre N. Safiullin 
//     gives permission to link the code of this program with
//     the JECL (Jabber External Component Library, jabber.tigris.org)
//     (or with modified versions of JECL that use the
//     same license as JECL), and distribute linked combinations including
//     the two.  You must obey the GNU General Public License in all
//     respects for all of the code used other than JECL.  If you modify
//     this file, you may extend this exception to your version of the
//     file, but you are not obligated to do so.  If you do not wish to
//     do so, delete this exception statement from your version.

/* File: MySocketWatcher.h -> a part of the "icqv7-t" project 
   Created: Thu Nov 15 19:35:18 2001
		by Alexandre N. Safiullin <alex_@unis-ru.com>
   Contains: asychronous selecter on sockets.  */

#ifndef MySocketWatcher_h_
#define MySocketWatcher_h_

#include <bedrock.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_EXT_HASH_MAP
# include <ext/hash_map>
#elif defined HAVE_HASH_MAP
# include <hash_map>
#else
# error "hash_map not defined"
#endif

using namespace bedrock;
using namespace std;
using namespace bedrock::net;

class it_Client;
class MySocketWatcher  {
public:
	MySocketWatcher(ThreadPool& pool, unsigned int maxsockets=0); //0 means unlimited
	~MySocketWatcher();

	void addSocket(int fd, it_Client *, int mode);
	void removeSocket(int fd);

	static void* beginProcessing(void* arg);

	struct exception
	{
		class WatcherFull {};
	};

private:

	void process(void);

	bool                 _started;          // whether selecter is started

	unsigned int         _socket_cnt;       // Max # of sockets
	ThreadPool&          _threadpool;
	pthread_t            _watcher_thread;   // Processing thread ID
	Mutex                _lock;             // Lock to protect 
	// non-constant members

	// Non-constant members (protected by _lock)
	hash_map<int, it_Client*> sessions;

    // descriptors for select
	fd_set               _read_set;              
	fd_set               _write_set;             
    
};



#endif
