/**
 * 
 * Copyright (C) 2001,2002 Alexandre. N. Safiullin <alex_@unislabs.com> 
 * 
 * The contents of this file are subject to the GNU General Public License
 * Version 2 (the License). You may not copy or use this
 * file, in either source code or executable form, except in compliance
 * with the License.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

//     In addition, as a special exception, Alexandre N. Safiullin 
//     gives permission to link the code of this program with
//     the JECL (Jabber External Component Library, jabber.tigris.org)
//     (or with modified versions of JECL that use the
//     same license as JECL), and distribute linked combinations including
//     the two.  You must obey the GNU General Public License in all
//     respects for all of the code used other than JECL.  If you modify
//     this file, you may extend this exception to your version of the
//     file, but you are not obligated to do so.  If you do not wish to
//     do so, delete this exception statement from your version.


/* File: it_Client.cpp -> a part of the "ICQv7-t" project 
   Created: Fri Nov  9 00:56:24 2001
		by Alexandre N. Safiullin <alex_@unis-ru.com>
   Contains: implementation of it_Client */

#include <it_Client.h>

// void it_Client::PPDB::addResource(const string& jid, MyElement* me){
// 	int prio=jutil_priority(me->jp->x);
// 	if(max_prio<prio)max_prio=prio;
// 	resources[jid]=prio;
// }

// list<const string> it_Client::PPDB::getMaxPrioResources(){
// 	list<const string> r;
// 	typedef hash_map<const string,int,_HashString>::iterator it;
// 	for(it i=resources.begin();i!=resources.end();i++)
// 		if(i->second==max_prio)r.insert(r.begin(),i->first);
// 	return r;
// }

it_Client::it_Client(IcqTrans& _icq_t, const string& _id, const unsigned int uin, 
					 const string& password) :  
	Client(uin,password), icq_t(_icq_t),
	confirmed(false), pinger(this), last_requested(STATUS_ONLINE),
	reconnecting(false), disconnect_key(0), exit_flag(false), id(_id) {

	contacts.setDefaultTimeout(300); // 5min between fetching

	icq_t.sessions[id]=this;

	logger.connect(slot(this,&it_Client::logger_cb));
	connected.connect(slot(this,&it_Client::connected_cb));
	contact_status_change_signal.connect(slot(this,&it_Client::contact_sc_cb));
	messaged.connect(slot(this,&it_Client::messaged_cb));
	disconnected.connect(slot(this,&it_Client::disconnected_cb));
	socket.connect(slot(this,&it_Client::socket_cb));
	want_auto_resp.connect(slot(this,&it_Client::want_auto_resp_cb));
	messageack.connect(slot(this,&it_Client::messageack_cb));
}

it_Client::~it_Client(){
  // TODO:
  // 1) Firstly disconnect

  // maybe by this way:
	pinger.stop();
	setStatus(STATUS_OFFLINE);
}

void it_Client::logger_cb(LogEvent *c){
	switch(c->getType()) {
	case LogEvent::INFO:
		t<<c->getMessage()<<endl<<flush;
		//	log_debug(ZONE,"%s\n",c->getMessage().c_str());
		break;
	case LogEvent::WARN:
		t<<c->getMessage()<<endl<<flush;
		//	log_warn(ti->i->id,"%s\n",c->getMessage().c_str());
		break;
	case LogEvent::PACKET:
	case LogEvent::DIRECTPACKET:
		//#ifdef PACKET_DEBUG
//		t << "[32m";	
		t << c->getMessage() << endl;
//		t << "[39m"<<flush;
		//#endif
		break;
	case LogEvent::ERROR:
//		t << "[31m :-<  ";
		t << c->getMessage() << endl;
//		t << "[39m";
		break;
	}
}

void it_Client::connected_cb(ConnectedEvent* ce)
{
	t<<"Connected!!"<<endl<<flush;
	
	m.lock();
	if (exit_flag){
		m.unlock();
		return;
    }
	m.unlock();

	icq_t.confirmConnection(id);
	
/* FIXME: hang up is possible here (may be) */
	while(!nonhandled.empty()){
		packetReceived(nonhandled.front());
		nonhandled.pop();
	}

	pinger.start();

	icq_t.sendSmsContacts(id);
	broadcastSmsPresence(true);
}

void it_Client::connect(){
	if(getStatus()==STATUS_OFFLINE)setStatus(last_requested);
// 	int old_fd=sock_fd;

// 	sock_fd=Connect();

// 	if(sock_fd!=-1&&old_fd!=sock_fd)
// 		icq_t.selecter.addSocket(sock_fd,this);
// 	else
// 		t<<"connect returns -1 || already connected"<<endl;
}

void it_Client::start(){

	/* maybe the internal part will send to us the contact list 
	   before we've got connected */
	icq_t.timer.schedule(3,&icq_t._tpool,icq_t._threadkey,
				   bedrock::callback::wrap(this,&it_Client::connect),1);
}

void  it_Client::packetReceived(MyElement* me){
	jpacket jp=me->jp;


	m.lock();
	
	if (exit_flag)
	{
		if (jp->type != JPACKET_PRESENCE)
		{
			t<<"Internal error, we shouldn't received it :( Let's ignore"<<endl;
			delete me;
		}
		else
			delete me;

		t<<"!! exit_flag is true, packet ignored!"<<endl;
		m.unlock();
		return;
	}

	m.unlock();

    switch (jp->type)
    {
	case JPACKET_IQ:
        iq(me);
        break;
    case JPACKET_PRESENCE:
        presence(me);
        break;
    case JPACKET_MESSAGE:
        message(me);
        break;
    case JPACKET_S10N:
        s10n(me);
        break;
    default:
		delete me;
    }
}
  
void it_Client::iq(MyElement * me){
	if(icq_t._sms_id==string(me->jp->to->server)){
		delete me;
		return;
	}

    jpacket jp=me->jp;
	if(!isConnected()){
		// postpone handling until connection
		MutexProxy mp(m);
		nonhandled.push(me);
		return;
	}

	char* ns=xmlnode_get_attrib(jp->iq,"xmlns");

	switch (me->subtype)
    {
    case JPACKET__GET:
		if (j_strcmp(ns,NS_AGENT)==0      ||
				 j_strcmp(ns,NS_REGISTER) == 0 ||
				 j_strcmp(ns,NS_VERSION) == 0  || 
				 j_strcmp(ns,NS_TIME) == 0     ||
				 j_strcmp(ns,NS_GATEWAY) == 0  ||
				 j_strcmp(ns,NS_ADMIN) == 0    ||
				 j_strcmp(ns,NS_BROWSE) == 0){
			t<<"Internal error, we shouldn't received it :( Let's ignore"<<endl;
			delete me;
		}

        else if (j_strcmp(ns,NS_SEARCH) == 0){
            jutil_error(jp->x,TERROR_NOTIMPL);
            icq_t.deliver(me,true);
			return;
        }
		
		else if(j_strcmp(ns,NS_VCARD) == 0 )vcard(me);

		else if(j_strcmp(ns,NS_LAST) == 0){
			unsigned int uin=utils::string2ui(me->jp->to->user);
			if(!uin){
				jutil_error(jp->x,(terror){400,"Bad uin"});
				icq_t.deliver(me,true);
				return;
			}
			ContactRef c=getContact(uin);
			
			if(c.get()&&c->getStatus()==STATUS_OFFLINE&&
			   c->get_last_online_time()){

				char str[16];
				xmlnode q,x;
				x=jutil_iqresult(jp->x);
				
				snprintf(str,16,"%d",time(NULL)-c->get_last_online_time());
				q = xmlnode_insert_tag(x,"query");
				xmlnode_put_attrib(q,"xmlns",NS_LAST);
				xmlnode_put_attrib(q,"seconds",str);
				
				icq_t.deliver(me,true);
			}
			else delete me;

		}
			
        break;

    case JPACKET__SET:
		if (j_strcmp(ns,NS_REGISTER) == 0 || j_strcmp(ns,NS_SEARCH) == 0){
			t<<"delivering error!"<<endl;
            jutil_error(jp->x,TERROR_NOTIMPL);
            icq_t.deliver(me,true);
		}
		else {
			t<<"Internal error, we shouldn't received it :( Let's ignore"<<endl;
			delete me;
		}
        break;

    default:
		delete me;
        break;
    }
}

void it_Client::disconnect(){
	m.lock();
	if(reconnecting&&!exit_flag){
		t<<"unscheduling reconnect"<<endl;
		icq_t.timer.unschedule(reconnect_key);
		exit_flag=true;
		m.unlock();
		
		icq_t.timer.schedule(1, /* 1 sec to wait */
							 &icq_t._tpool,icq_t._threadkey,
							 bedrock::callback::wrap(&icq_t,&IcqTrans::end,id),1); 
	}
	else {
		m.unlock();
		setStatus(STATUS_OFFLINE);
	}
}

void it_Client::presence(MyElement* me)
{
	jpacket jp=me->jp;
    switch (jpacket_subtype(jp))
    {
    case JPACKET__PROBE:
        if (jp->to->user)
        {
			if(icq_t._sms_id==string(jp->to->server)){
				if(!isConnected())
					sms_contacts.insert(jp->to->user);
				else 
					sendSmsPresence(jp->to->user);
			}
			else
				sendContactPresence(utils::string2ui(jp->to->user));
        }
        break;

    case JPACKET__AVAILABLE:
    {
        if (jp->to->user){
			unsigned int uin=utils::string2ui(jp->to->user);
			if(getInvisible()){
				ContactRef c=getContact(uin);
				if(!c.get())c=ContactRef(new Contact(uin));
				addVisible(c);
			}
			else removeInvisible(uin);
            break;
		}

        last_requested = utils::show2status(xmlnode_get_tag_data(jp->x,"show"));

		char* text=xmlnode_get_tag_data(jp->x,"status");
		status_text=text?text:"";


		setStatus(last_requested);
		setInvisible(false);

		if(!isConnected())
			connect();
		break;
    }

    case JPACKET__UNAVAILABLE:
		if (jp->to->user == NULL)
        { 
			m.lock();
			disconnect_key=icq_t.timer.schedule(2, /* 2 sec to wait invisible presence */
											   &icq_t._tpool,icq_t._threadkey,
											   bedrock::callback::wrap(this,&it_Client::disconnect),1); 
			m.unlock();


// it will be called in disconnected_cb
// 			icq_t.end(id);
		}
		break;


	case JPACKET__INVISIBLE:
        if (jp->to->user){
			unsigned int uin=utils::string2ui(jp->to->user);
			ContactRef c=getContact(uin);
			if(!c.get())c=ContactRef(new Contact(uin));
			if(getInvisible())
				removeVisible(uin);
			else addInvisible(c);
            break;
		}
		else {
			m.lock();
			if(disconnect_key){
				t<<"unscheduling shutdown"<<endl;
				icq_t.timer.unschedule(disconnect_key);
				disconnect_key=0;
				/* need to confirm */
				icq_t.confirmConnection(id);
			}
			else t<<"disconnect key is 0"<<endl;
			m.unlock();
			setInvisible(true);
		}
		break;
    default:
        break;
    }

	delete me;
}

void it_Client::message(MyElement* me){
	jpacket jp=me->jp;
	xmlnode x;
	ContactRef c;

	if(!jp->to->user){
		jutil_error(jp->x,TERROR_BAD);
		icq_t.deliver(me,true);
		return;
	}

	unsigned int uin=utils::string2ui(jp->to->user);
	t<<"it_Client::message for user: "<<uin<<endl;

	if(uin==0){
		jutil_error(jp->x,(terror){400,"Bad uin"});
		icq_t.deliver(me,true);
		return;
	}
	else {
		if(icq_t._sms_id==string(jp->to->server)){
			ContactRef nc(new Contact());
			nc->setMobileNo(jp->to->user);
			sendSMS(nc,me);
			return;
		}
		c=getContact(uin);
		if(c.get()==NULL)
			c=ContactRef(new Contact(uin));
// 			ContactRef nc(new Contact(uin));
// 			addContact(nc);
// 			c=getContact(uin);

		
	}

	if ((x = xmlnode_get_tag(jp->x,"x?xmlns=jabber:x:roster"))){
		jutil_error(jp->x,TERROR_NOTIMPL);
		icq_t.deliver(me,true);
	}
    else if ((x = xmlnode_get_tag(jp->x,"x?xmlns=jabber:x:oob")) != NULL)
		sendURL(c,me,x);
	else
        sendText(c,me);
}

void it_Client::sendSMS(ContactRef c, MyElement *me){
	jpacket jp=me->jp;
	char* body= xmlnode_get_tag_data(jp->x,"body");

	if(body==NULL){
		jutil_error(jp->x,
					(terror){400,"Empty messages are forbidden"});
		icq_t.deliver(me,true);
		return;
	}

	if(strlen(body)>=SMS_Max_Length){
		jutil_error(jp->x,
					(terror){400,"Too long message, the limit is 160 chars"});
		icq_t.deliver(me,true);
		return;
	}
// 	string unutfed=utils::_unUTF8(string(body));
// 	if(unutfed.size()!=strlen(body)){
// 		jutil_error(jp->x,
// 					(terror){400,"Only ASCII accepted"});
// 		icq_t.deliver(me,true);
// 		return;
// 	}

	SMSMessageEvent*  message=new SMSMessageEvent(c,body,true);
	SendEvent(message);

   
	nonacked[(unsigned long)message]=me;
//	delete me;
}

void it_Client::sendURL(ContactRef c, MyElement* me, xmlnode x){
	char* url=xmlnode_get_tag_data(x,"url");
	char* desc=xmlnode_get_tag_data(x,"desc");
	jpacket jp=me->jp;

	if(url==NULL){
		jutil_error(jp->x,
					(terror){400,"Empty urls are forbidden"});
		icq_t.deliver(me,true);
		return;
	}

	
	string description=utils::_unUTF8(string(desc?desc:""));
	string smart_url=utils::_unUTF8(string(url));

	if((desc&&!description.size())||
	    !smart_url.size()){
		jutil_error(jp->x,
					(terror){400,"Bad UTF8"});
		icq_t.deliver(me,true);
		return;
	}

	URLMessageEvent* urlmessage=new URLMessageEvent(c,description,smart_url);

	if(c->getStatus()==STATUS_DND||c->getStatus()==STATUS_OCCUPIED)
		urlmessage->setUrgent(true);
	SendEvent(urlmessage);

	nonacked[(unsigned long)urlmessage]=me;
//	delete me;
}

void it_Client::sendText(ContactRef c, MyElement* me){
	jpacket jp=me->jp;
	char* body= xmlnode_get_tag_data(jp->x,"body");

	if(body==NULL){
		jutil_error(jp->x,
					(terror){400,"Empty messages are forbidden"});
		icq_t.deliver(me,true);
		return;
	}
	char* type_attrib=xmlnode_get_attrib(jp->x,"type");

	chat[c->getUIN()] = (type_attrib==NULL||string(type_attrib)!="chat")?false:true;

	const string unutfed=utils::_unUTF8(string(body));
	if(strlen(body)&&!unutfed.size()){
		jutil_error(jp->x,
					(terror){400,"Bad UTF8"});
		icq_t.deliver(me,true);
		return;
	}

	NormalMessageEvent*  message=new NormalMessageEvent(c,unutfed);
	if(c->getStatus()==STATUS_DND||c->getStatus()==STATUS_OCCUPIED)
		message->setUrgent(true);
	SendEvent(message);

	nonacked[(unsigned long)message]=me;
//	delete me;
}

void it_Client::s10n(MyElement* me){

	unsigned int uin=utils::string2ui(me->jp->to->user);
	t<<"it_Client::s10n for user: "<<uin<<endl;

	if(!uin){
		delete me;
		return;
	}
	
	if(icq_t._sms_id==string(me->jp->to->server)&&
		me->subtype!=JPACKET__SUBSCRIBE&&
		me->subtype!=JPACKET__UNSUBSCRIBE){
		delete me;
		return;
	}

	switch(me->subtype){
	case JPACKET__SUBSCRIBE:
	{ 
		if(icq_t._sms_id==string(me->jp->to->server)){
			string mobile_no=me->jp->to->user;
			sms_contacts.insert(mobile_no);

			jutil_tofrom(me->jp->x);
			xmlnode_put_attrib(me->jp->x,"type","subscribed");

			icq_t.deliver(me);

			sendSmsPresence(mobile_no);
			return;
		}
		ContactRef c=getContact(uin);
		if(c.get()==NULL){
			addContact(ContactRef(new Contact(uin)));
			t<<"Contact added: "<<uin<<flush<<endl;
		}
		if(isConnected())sendContactPresence(uin);
		break;
	}
	case JPACKET__SUBSCRIBED:
		authorizeUser(uin,true);
		break;
	case JPACKET__UNSUBSCRIBED:
		authorizeUser(uin,false);
		break;
	case JPACKET__UNSUBSCRIBE:
		if(icq_t._sms_id==string(me->jp->to->server)){
			sms_contacts.erase(me->jp->to->user);
		}
		else{ 
			removeContact(uin);
			chat.erase(uin);
		}
		break;
	default:
		break;
	}
	delete me;
}
	
void it_Client::sendContactPresence(const unsigned int uin, const string& status){

	ContactRef c=getContact(uin);
	if(c.get()==NULL)
		return;
	
	Status st=c->getStatus();
	xmlnode x, pres = jutil_presnew(st==STATUS_OFFLINE ? JPACKET__UNAVAILABLE : JPACKET__AVAILABLE,id.c_str(),status.size()==0?NULL:status.c_str());

    xmlnode_put_attrib(pres,"from",(utils::ui2string(uin)+
									"@"+icq_t.getInternalName()).c_str());

    switch (st)
    {
    case STATUS_AWAY:
        x = xmlnode_insert_tag(pres,"show");
        xmlnode_insert_cdata(x,"away",-1);
        break;
    case STATUS_DND:
        x = xmlnode_insert_tag(pres,"show");
        xmlnode_insert_cdata(x,"dnd",-1);
        break;
    case STATUS_NA:
        x = xmlnode_insert_tag(pres,"show");
        xmlnode_insert_cdata(x,"xa",-1);
        break;
    case STATUS_OCCUPIED:
        x = xmlnode_insert_tag(pres,"show");
        xmlnode_insert_cdata(x,"dnd",-1);
        break;
    case STATUS_FREEFORCHAT:
        x = xmlnode_insert_tag(pres,"show");
        xmlnode_insert_cdata(x,"chat",-1);
        break;
    default:
		break;
    }
    icq_t.deliver(new MyElement(pres));
}

void it_Client::sendSmsPresence(const string& mobile_no, bool available){

	t<<"sending SMS presence for "<<mobile_no<<endl;

	xmlnode x, pres = jutil_presnew((available && icq_t._sms_show.size()) ? JPACKET__AVAILABLE : JPACKET__UNAVAILABLE,id.c_str(),NULL);

    xmlnode_put_attrib(pres,"from",(mobile_no+"@"+icq_t._sms_id).c_str());

	if(icq_t._sms_show.size()&&available&&icq_t._sms_show!="online"){
		x = xmlnode_insert_tag(pres,"show");
		xmlnode_insert_cdata(x,icq_t._sms_show.c_str(),icq_t._sms_show.size());
	}

    icq_t.deliver(new MyElement(pres));
}

void it_Client::broadcastSmsPresence(bool available){
	t<<"broadcastingSmsPresence -- "<<endl;
	for(set<string>::const_iterator ci=sms_contacts.begin();
		ci!=sms_contacts.end(); ci++)
		sendSmsPresence(*ci, available);
}

void it_Client::contact_sc_cb(StatusChangeEvent *ev){
	ContactRef c=ev->getContact();
	t<<"Contact "<<c->getUIN()<<" changed status"<<endl;
	sendContactPresence(c->getUIN());
}

void it_Client::messageack_cb(MessageEvent* ev){
	if(!ev->isFinished())return;

	ContactRef c=ev->getContact();
	const unsigned int uin=c->getUIN();
	MyElement* me=nonacked.find((unsigned long)ev) != nonacked.end() ? 
		nonacked[(unsigned long)ev] : NULL;
	
	nonacked.erase((unsigned long)ev);

	if (ev->getType() == ICQ2000::MessageEvent::AwayMessage){
		AwayMessageEvent *aev = static_cast<AwayMessageEvent*>(ev);
		t<<"Away message received"<<endl;
		sendContactPresence(uin,utils::_2UTF8(aev->getAwayMessage()));
		if(me)delete me;
		return;
	}

	if(!ev->isDelivered()){
		if(!me)
			t<<"Unsaved packet acked as undelivered"<<endl;
		else {
			switch(ev->getDeliveryFailureReason()) {
				
			case MessageEvent::Failed_Denied:
				jutil_error(me->jp->x,
							(terror){502,"Sending message failed - user is ignoring you"});
				break;
				
			case MessageEvent::Failed_Occupied:
				jutil_error(me->jp->x,
							(terror){502,"Sending message failed - user is occupied"});
				break;
			case MessageEvent::Failed_DND:
				jutil_error(me->jp->x,
							(terror){502,"Sending message failed - user is in do not disturb"});
				break;
			case MessageEvent::Failed_NotConnected:
				jutil_error(me->jp->x,
							(terror){502,"Sending message failed - you are not connected"});
				break;
			case MessageEvent::Failed:
			default:
				jutil_error(me->jp->x,
							(terror){502,"Sending message failed"});
			}
			xmlnode_put_attrib(me->jp->x,"from",(utils::ui2string(uin)+
										   "@"+icq_t.getInternalName()).c_str());
			icq_t.deliver(me);
			
		}
	}
	else if(me)delete me;
	
	ICQMessageEvent *cev = dynamic_cast<ICQMessageEvent*>(ev);
	if(cev&&cev->getAwayMessage().size())
		sendContactPresence(uin,utils::_2UTF8(cev->getAwayMessage()));
}

void it_Client::messaged_cb(MessageEvent* m){
	
	ContactRef c=m->getContact();
	const unsigned int uin=c->getUIN();
	xmlnode msg;
	switch(m->getType()){
	case MessageEvent::Normal: {
		NormalMessageEvent* nm=static_cast<NormalMessageEvent*>(m);
		msg = xmlnode_new_tag("message");
		bool type_chat = chat.find(uin) != chat.end() ? chat[uin] : false;

		if(type_chat)xmlnode_put_attrib(msg,"type","chat");

		xmlnode_insert_cdata(xmlnode_insert_tag(msg,"body"),
							 utils::_2UTF8(nm->getMessage()).c_str(),-1);

		break;
	}
	case MessageEvent::URL: {
		URLMessageEvent* urlm=static_cast<URLMessageEvent*>(m);
	    msg = xmlnode_new_tag("message");

		xmlnode_insert_cdata(xmlnode_insert_tag(msg,"subject"),"ICQ URL Message",-1);

		/* add a message body that most clients can understand */
		xmlnode_insert_cdata(xmlnode_insert_tag(msg,"body"),
							 ("URL: "+utils::_2UTF8(urlm->getURL())+"\nDescription:\n"+
							  utils::_2UTF8(urlm->getMessage())).c_str(),-1);

		xmlnode x = xmlnode_insert_tag(msg,"x");
		xmlnode_put_attrib(x,"xmlns",NS_XOOB);
		xmlnode_insert_cdata(xmlnode_insert_tag(x,"desc"),
							 utils::_2UTF8(urlm->getMessage()).c_str(),-1);
		xmlnode_insert_cdata(xmlnode_insert_tag(x,"url"),utils::_2UTF8(urlm->getURL()).c_str(),-1);
		break;
	}
	case MessageEvent::AuthReq: {
		t<<"Authorization request received"<<endl;
		msg = xmlnode_new_tag("presence");
		xmlnode_put_attrib(msg,"type","subscribe");
		break;
	}
	case MessageEvent::AuthAck: {
		t<<"Authorization received"<<endl;
		AuthAckEvent* ackm=static_cast<AuthAckEvent*>(m);
		msg = xmlnode_new_tag("presence");
		if(ackm->isGranted())
			xmlnode_put_attrib(msg,"type","subscribed");
		else 
			xmlnode_put_attrib(msg,"type","unsubscribed");
		break;
	}
	case MessageEvent::SMS: {
		t<<"SMS received"<<endl;
		SMSMessageEvent* smsm=static_cast<SMSMessageEvent*>(m);
		
		msg = xmlnode_new_tag("message");
		string text=smsm->getMessage();

		xmlnode_insert_cdata(xmlnode_insert_tag(msg,"body"),
							 text.c_str(),text.size());
			
		xmlnode_put_attrib(msg,"to",id.c_str());
		xmlnode_put_attrib(msg,"from",(c->getMobileNo()+
									   "@"+icq_t._sms_id).c_str());
		icq_t.deliver(new MyElement(msg));
		m->setDelivered(true);
		return;
	}
	case MessageEvent::SMS_Receipt: {
		t<<"SMS receipt received"<<endl;
		SMSReceiptEvent* receipt=static_cast<SMSReceiptEvent*>(m);
		
		msg = xmlnode_new_tag("message");
		string text= "Your message "+ 
			( receipt->delivered() ? 
			  string("has been delivered on ") + receipt->getDeliveryTime() :
			  "failed to delivery" );

		xmlnode_insert_cdata(xmlnode_insert_tag(msg,"body"),
							 text.c_str(),text.size());
			
		xmlnode_put_attrib(msg,"to",id.c_str());
		xmlnode_put_attrib(msg,"from",(c->getMobileNo()+
									   "@"+icq_t._sms_id).c_str());
		icq_t.deliver(new MyElement(msg));
		m->setDelivered(true);
		return;
	}
	case MessageEvent::UserAdd: {
		t<<"Somebody added you"<<endl;
//		UserAddEvent* receipt=static_cast<UserAddEvent*>(m);
		msg = xmlnode_new_tag("message");
		xmlnode_insert_cdata(xmlnode_insert_tag(msg,"body"),
							 "ICQv7-t: The user added you, enjoy chatting",-1);
		break;
	}
	case MessageEvent::AwayMessage: {
		t<<"Away message received"<<endl;
		AwayMessageEvent *aev = static_cast<AwayMessageEvent*>(m);
			
		sendContactPresence(aev->getContact()->getUIN(),utils::_2UTF8(aev->getAwayMessage()));
		return;
	}
	case MessageEvent::EmailEx: {
		t<<"email express received"<<endl;
		EmailExEvent* eeev=static_cast<EmailExEvent*>(m);
		
		msg = xmlnode_new_tag("message");
		xmlnode_insert_cdata(xmlnode_insert_tag(msg,"body"),
							 (utils::_2UTF8(eeev->getSender())+" on behalf of "+
							  utils::_2UTF8(eeev->getEmail())+
							  " sent you the following message:\n"+
							  utils::_2UTF8(eeev->getMessage())).c_str(),-1);
		break;
	}
	case MessageEvent::Email: {
		t<<"email message received"<<endl;
		EmailMessageEvent* emev=static_cast<EmailMessageEvent*>(m);
		
		msg = xmlnode_new_tag("message");
		xmlnode_insert_cdata(xmlnode_insert_tag(msg,"body"),
							 (utils::_2UTF8(emev->getMessage())).c_str(),-1);

		break;
	}
// 	default:
// 		msg = xmlnode_new_tag("message");
// 		char s[]="You receive some SMS related stuff, but the ICQv7-t cannot say exactly what.\n\nWait till SMS-subtransport will be implemented.";
// 		xmlnode_insert_cdata(xmlnode_insert_tag(msg,"body"),s,-1);
// 		break;
	}
	
	xmlnode_put_attrib(msg,"to",id.c_str());
	xmlnode_put_attrib(msg,"from",(utils::ui2string(uin)+
								   "@"+icq_t.getInternalName()).c_str());
	icq_t.deliver(new MyElement(msg));

	m->setDelivered(true);

}

void it_Client::disconnected_cb(DisconnectedEvent* de){
	t<<"disconnected_cb"<<endl;

	bool fatal=false;
	bool toconfirm=true;

	string error_msg;
	int error_code;

	switch(de->getReason()){
	case DisconnectedEvent::FAILED_BADUSERNAME:
		error_msg="Bad username";
		error_code=400;
		fatal=true;
		break;
	case DisconnectedEvent::FAILED_TURBOING:
		error_msg="Turboing, connect later";
		error_code=503;
		break;
	case DisconnectedEvent::FAILED_BADPASSWORD:
		error_msg="Bad (non mismatched) registration password";
		error_code=400;
		fatal=true;
		break;
	case DisconnectedEvent::FAILED_MISMATCH_PASSWD:
		error_msg="Password doesn't match";
		error_code=401;
		fatal=true;
		break;
	case DisconnectedEvent::FAILED_DUALLOGIN: {
		/* maybe we should explicitly notify due importance */
		char *s=strdup(id.c_str());
		char *body="Your UIN is used by somebody else. Dual login. Reconnect disabled";
		xmlnode msg = xmlnode_new_tag("message");
		xmlnode_put_attrib (msg, "to", s);
		xmlnode_insert_cdata (xmlnode_insert_tag (msg, "body"), body, strlen (body));
		xmlnode_put_attrib(msg,"from",icq_t.getInternalName().c_str());

		free(s);

		icq_t.deliver(new MyElement(msg));

		error_msg="Dual login";
		error_code=409;
		fatal=true;
		break;
	}
	case DisconnectedEvent::FAILED_LOWLEVEL:
		error_msg="Low level network error";
		error_code=502;
		break;
	case DisconnectedEvent::FAILED_UNKNOWN:
		error_msg="Disconnected by unknown reason";
		error_code=502;
		break;
	case DisconnectedEvent::REQUESTED:
		t<<"Disconnected on request"<<endl;
		toconfirm=false; //they know it ;)
		fatal=true;
		break;

	}

	pinger.stop();
	broadcastSmsPresence(false);

// 	if(sock_fd!=-1)icq_t.selecter.removeSocket(sock_fd);
// 	t<<"Socket is taken away"<<endl<<flush;
	
	
	if(!fatal && icq_t.reconnect){
		t<<"Scheduling reconnect"<<endl;
		m.lock();
		reconnecting=true;
		reconnect_key=icq_t.timer.schedule(60, /* 60 sec to wait */
					   &icq_t._tpool,icq_t._threadkey,
					   bedrock::callback::wrap(this,&it_Client::reconnect),1); 
		m.unlock();
	}
	else
		fatal=true;

	if(toconfirm)icq_t.timer.schedule(1, /* 1 sec to wait */
									  &icq_t._tpool,icq_t._threadkey,
			  bedrock::callback::wrap(&icq_t,
									  &IcqTrans::confirmDisconnect,
									  id,
									  error_msg,
									  error_code,
									  fatal),
									  1);
	if(fatal){
		MutexProxy mp(m);
		exit_flag=true;

		icq_t.timer.schedule(1, /* 1 sec to wait */
							 &icq_t._tpool,icq_t._threadkey,
							 bedrock::callback::wrap(&icq_t,&IcqTrans::end,id),1); 
	}
// 	sock_fd=-1;
	t<<"disconnect_cb finished"<<endl;
}


void it_Client::reconnect() {
	m.lock();
	reconnecting=false;
	m.unlock();
	connect();
}

void it_Client::socket_cb(SocketEvent* ev){
	/* I liked old good case(se->getType()) more */

	t<<"socket_cb triggered"<<endl;
	if (dynamic_cast<AddSocketHandleEvent*>(ev) != NULL) {
		AddSocketHandleEvent *cev = dynamic_cast<AddSocketHandleEvent*>(ev);
		int fd = cev->getSocketHandle();
		int mode=0;
		if(cev->isException()){
			t<<"Fearful socket for exception arrived! I'm scared, what to do?"<<endl;
			return;
		}
		if(cev->isRead())mode|=SocketEvent::READ;
		if(cev->isWrite())mode|=SocketEvent::WRITE;

		t<<"Adding socket "<<fd<<" with mode: "<<mode<<endl;
		if(mode!=0)icq_t.selecter.addSocket(fd,this,mode);

	} else if (dynamic_cast<RemoveSocketHandleEvent*>(ev) != NULL) {
		RemoveSocketHandleEvent *cev = dynamic_cast<RemoveSocketHandleEvent*>(ev);
		int fd = cev->getSocketHandle();
		if(fd!=-1)icq_t.selecter.removeSocket(fd);
		else
			t<<"Got -1 descriptor!"<<endl;
    }
}

void it_Client::vcard(MyElement* me){
	
	MutexProxy mp(m); // lock contacts

	unsigned int uin=utils::string2ui(me->jp->to->user);
	time_t now=time(NULL);

	if(contacts.exists(uin)){
		time_t last_fetch=contacts.getTimestamp(uin);

		if(now-last_fetch<3) // let's await info arrival
			contacts[uin]->enqueue(me);
		else
			sendVcard(me,contacts[uin]->c);
	}
	else {
		ContactRef c=getContact(uin);
		if(c.get()!=NULL)
			contacts.insert(uin, new ContactInfoRequest(c))->enqueue(me);
		else{
			c=ContactRef(new Contact(uin));
			contacts.insert(uin, new ContactInfoRequest(c))->enqueue(me);
		}
		fetchDetailContactInfo(c);
		
		if(c->getStatus()!=STATUS_ONLINE&&
		   c->getStatus()!=STATUS_OFFLINE){
// the place to request away message for contact

			MessageEvent *ev = new AwayMessageEvent(c);
			SendEvent(ev);
		}

		icq_t.timer.schedule(3, /* 3 sec to wait info arrival */
							 &icq_t._tpool,icq_t._threadkey,
							 bedrock::callback::wrap(this,
													 &it_Client::handleInfoRequests,uin),1); 
	}
}

void it_Client::handleInfoRequests(unsigned int uin){
	MutexProxy mp(m); //lock contacts

	if(!contacts.exists(uin)){
		t<<"Internal error, no contact in cache!"<<endl;
		return;
	}
	queue<MyElement*>& q=contacts[uin]->nonhandled;
	ContactRef c=contacts[uin]->c;
	while(!q.empty()){
		sendVcard(q.front(),c);
		q.pop();
	}
}

#define VCARD_ADD(qq, parent, name, value) \
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_insert_tag(qq,parent),name), \
    value.c_str(),value.size())

void it_Client::sendVcard(MyElement* me, ContactRef c){
	/* contacts are locked by caller */
	unsigned int uin=utils::string2ui(me->jp->to->user);
	
	t<<"Sending vcard of "<<uin<<endl;

	xmlnode r=me->jp->x;
    jutil_iqresult(r);

	xmlnode q,temp;

// 	xmlnode_put_attrib(r,"type","result");
// 	xmlnode_put_attrib(r,"to",id.c_str());
 	xmlnode_put_attrib(r,"from",(utils::ui2string(uin)+"@"
								 +icq_t.getInternalName()).c_str());

    q = xmlnode_insert_tag(r,"vCard");
    xmlnode_put_attrib(q,"xmlns",NS_VCARD);
    xmlnode_put_attrib(q,"version","3.0");
    xmlnode_put_attrib(q,"prodid","-//HandGen//NONSGML vGen v1.0//EN");
	
	string cur=utils::_2UTF8(c->getFirstName()+" "+c->getLastName());
	
    if (cur.size()!=1)
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"FN"),
							 cur.c_str(),cur.size());
	
    temp = xmlnode_insert_tag(q,"N");

/* First Name : */
	cur=utils::_2UTF8(c->getFirstName());
	
    if (cur.size())
        xmlnode_insert_cdata(xmlnode_insert_tag(temp,"GIVEN"),
							 cur.c_str(),cur.size());

/* Last Name : */
	cur=utils::_2UTF8(c->getLastName());
	
    if (cur.size())
        xmlnode_insert_cdata(xmlnode_insert_tag(temp,"FAMILY"),
							 cur.c_str(),cur.size());

/* Last Name : */
	cur=utils::_2UTF8(c->getAlias());
	
    if (cur.size())
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"NICKNAME"),
							 cur.c_str(),cur.size());

	cur=utils::_2UTF8(c->getEmail());
	
    if (cur.size())
    {
        temp = VCARD_ADD(q,"EMAIL","USERID",cur);
        xmlnode_insert_tag(temp,"INTERNET");
        xmlnode_insert_tag(temp,"PREF");
    }

//     if (info->email2)
//     {
//         temp = VCARD_ADD(q,"EMAIL","USERID",info->email2);
//         xmlnode_insert_tag(temp,"INTERNET");    
//     }

	cur=utils::_2UTF8(c->getMainHomeInfo().phone);
	
    if (cur.size())
    {
        temp = VCARD_ADD(q,"TEL","NUMBER",cur);
        xmlnode_insert_tag(temp,"VOICE");
        xmlnode_insert_tag(temp,"HOME");
    }

	cur=utils::_2UTF8(c->getMainHomeInfo().fax);
	
    if (cur.size())
    {
        temp = VCARD_ADD(q,"TEL","NUMBER",cur);
        xmlnode_insert_tag(temp,"FAX");
        xmlnode_insert_tag(temp,"HOME");
    }

	cur=utils::_2UTF8(c->getMobileNo());
	
    if (cur.size())
    {
        temp = VCARD_ADD(q,"TEL","NUMBER",cur);
        xmlnode_insert_tag(temp,"VOICE");
        xmlnode_insert_tag(temp,"CELL");   
    }

    temp = xmlnode_insert_tag(q,"ADR");
    xmlnode_insert_tag(temp,"HOME");
    xmlnode_insert_tag(temp,"EXTADD");

	cur=utils::_2UTF8(c->getMainHomeInfo().street);
	t<<"street:"<<cur<<endl;
    if (cur.size())
        xmlnode_insert_cdata(xmlnode_insert_tag(temp,"STREET"),
							 cur.c_str(),cur.size());

	cur=utils::_2UTF8(c->getMainHomeInfo().city);
	t<<"city:"<<cur<<endl;
    if (cur.size())
        xmlnode_insert_cdata(xmlnode_insert_tag(temp,"LOCALITY"),
							 cur.c_str(),cur.size());

	cur=utils::_2UTF8(c->getMainHomeInfo().state);
	
    if (cur.size())
        xmlnode_insert_cdata(xmlnode_insert_tag(temp,"REGION"),
							 cur.c_str(),cur.size());

	cur=utils::_2UTF8(c->getMainHomeInfo().zip);
	
    if (cur.size())
        xmlnode_insert_cdata(xmlnode_insert_tag(temp,"PCODE"),
							 cur.c_str(),cur.size());

	cur=utils::_2UTF8(c->getMainHomeInfo().getCountry());
	
    if (cur.size())
        xmlnode_insert_cdata(xmlnode_insert_tag(temp,"COUNTRY"),
							 cur.c_str(),cur.size());
	
	cur=utils::_2UTF8(c->getWorkInfo().company_name);
	
    if (cur.size())
    {
        temp = VCARD_ADD(q,"ORG","ORGNAME",cur);
		cur=utils::_2UTF8(c->getWorkInfo().company_dept);
		
        if (cur.size())
			xmlnode_insert_cdata(xmlnode_insert_tag(temp,"ORGUNIT"),
								 cur.c_str(),cur.size());
	}

//     if (work->phone)
//     {
//         temp = VCARD_ADD(q,"TEL","NUMBER",work->phone);
//         xmlnode_insert_tag(temp,"VOICE");
//         xmlnode_insert_tag(temp,"WORK");
//     }

//     if (work->fax)
//     {
//         temp = VCARD_ADD(q,"TEL","NUMBER",work->fax);
//         xmlnode_insert_tag(temp,"FAX");
//         xmlnode_insert_tag(temp,"WORK");
//     }

//     temp = xmlnode_insert_tag(q,"ADR");
//     xmlnode_insert_tag(temp,"WORK");
//     xmlnode_insert_tag(temp,"EXTADD");

// 	cur=utils::_2UTF8(c->getWorkInfo().street);
// 	if (cur.size())
//         xmlnode_insert_cdata(xmlnode_insert_tag(temp,"STREET"),
// 							 cur.c_str(),cur.size());

	
// 	cur=utils::_2UTF8(c->getWorkInfo().city);
// 	if (cur.size())
//         xmlnode_insert_cdata(xmlnode_insert_tag(temp,"LOCALITY"),
// 							 cur.c_str(),cur.size());

// 	cur=utils::_2UTF8(c->getWorkInfo().state);
// 	if (cur.size())
//         xmlnode_insert_cdata(xmlnode_insert_tag(temp,"REGION"),
// 							 cur.c_str(),cur.size());

// 	cur=utils::_2UTF8(c->getWorkInfo().zip);
// 	if (cur.size())
//         xmlnode_insert_cdata(xmlnode_insert_tag(temp,"PCODE"),
// 							 cur.c_str(),cur.size());

//     c = it_country_from_code(work->country_code);
//     xmlnode_insert_cdata(xmlnode_insert_tag(temp,"COUNTRY"),c,-1);

// 	cur=utils::_2UTF8(c->getWorkInfo().company_position);
// 	if (cur.size())
//         xmlnode_insert_cdata(xmlnode_insert_tag(q,"TITLE"),
// 							 cur.c_str(),cur.size());

	// ugh.. so much information, 
	// whar for?!
	// probably some fuckers would like to additionally specify favorite position of sex performing.. 

	// %-/

	
	cur=utils::_2UTF8(c->getHomepageInfo().homepage);
	
	if (cur.size())
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"URL"),
							 cur.c_str(),cur.size());

    /* birth day */
	cur=utils::_2UTF8(c->getHomepageInfo().getBirthDate());
	
	if (cur.size())
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"BDAY"),
							 cur.c_str(),cur.size());

	cur=utils::_2UTF8(c->getAboutInfo());

	cur+=(string("\n---added by ICQv7-t---\nIP: ")+IPtoString(c->getLanIP())
						   +":"+utils::ui2string(c->getLanPort()));

	if(c->getExtIP()!=c->getLanIP())cur+=(string("/")+IPtoString(c->getExtIP())
		  +":"+utils::ui2string(c->getExtPort()));
	cur+="\n";

//	if (cur.size())
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"DESC"),
							 cur.c_str(),cur.size());

	icq_t.deliver(me);
}

void it_Client::authorizeUser(unsigned int uin, bool granted){
	ContactRef c=getContact(uin);
	if(c.get()==NULL)
		c=ContactRef(new Contact(uin));
	AuthAckEvent* message=new AuthAckEvent(c,granted);
	SendEvent(message);
}

void it_Client::want_auto_resp_cb(ICQMessageEvent* ame){
	if(status_text.size()>0){
		ame->setAwayMessage(utils::_unUTF8(status_text));
	}
}
