/**
 * 
 * Copyright (C) 2001 Alexandre. N. Safiullin <alex_@unis-ru.com> 
 * 
 * The contents of this file are subject to the GNU General Public License
 * Version 2 (the License). You may not copy or use this
 * file, in either source code or executable form, except in compliance
 * with the License.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

//     In addition, as a special exception, Alexandre N. Safiullin 
//     gives permission to link the code of this program with
//     the JECL (Jabber External Component Library, jabber.tigris.org)
//     (or with modified versions of JECL that use the
//     same license as JECL), and distribute linked combinations including
//     the two.  You must obey the GNU General Public License in all
//     respects for all of the code used other than JECL.  If you modify
//     this file, you may extend this exception to your version of the
//     file, but you are not obligated to do so.  If you do not wish to
//     do so, delete this exception statement from your version.


/* File: it_Client.h -> a part of the "ICQv7-t" project 
   Created: Sun Nov  4 23:42:10 2001
		by Alexandre N. Safiullin <alex_@unislabs.com>
   Contains:  */

#ifndef it_Client_h_
#define it_Client_h_

#include <libicq2000/Client.h>
#include <IcqTrans.h>
#include <queue>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_ICONV
#include <iconv.h>
#else
#error "you need iconv to compile the package"
#endif



using namespace ICQ2000;

class it_Client : public Client {
private:
// 	int sock_fd; // socket to server
	IcqTrans& icq_t;

	Mutex m; // protects non-constant members below:
	queue<MyElement*> nonhandled;
	bool confirmed;

	class Pinger{
		pthread_t _thread;
		it_Client* session;

		static void* beginProcessing(void* arg); // static entrance point

		void process();

// no need
//		Mutex lock; 
		bool started;
	public:
		Pinger(it_Client* s): session(s), started(false) {};
		~Pinger(){if(started)stop();}
		void stop();
		void start();
	};

	Pinger pinger;

	struct ContactInfoRequest {
		ContactRef c;
		queue<MyElement*> nonhandled;
		void enqueue(MyElement* me){
			nonhandled.push(me);
		};
		ContactInfoRequest(ContactRef contact) : c(contact) {};
		~ContactInfoRequest(){
			/* FIXME: what about 'nonhandled' ? */
		}
	};
	
	class ContactsCache : public Cache<unsigned int,ContactInfoRequest*> {
		it_Client* session;
	public:
		void removeItem(const ContactsCache::literator& l);
		time_t getTimestamp(const unsigned int uin){
			return lookup(uin)->getTimestamp();
		}
	};

	ContactsCache contacts;

 /* uin vs message-type (chat,normal) 

    the point is if u send 'chat' to this uin, it will come 
    back as 'chat' too, and vice verse */

 	hash_map<unsigned int,bool> chat; 

	set<string> sms_contacts;

	Status last_requested;
	string status_text;

	bool reconnecting;
	int reconnect_key;
	int disconnect_key;

	/* MyElement* vs MessageEvent*  to track acks */
	hash_map<unsigned long,MyElement*> nonacked;

public:
	bool exit_flag;
	string id;
	
	
	it_Client(IcqTrans& _icq_t, const string& _id,
			  const unsigned int uin, const string& password);
	~it_Client();

	void start();
	void packetReceived(MyElement *);
private:	
/* Callbacks from libICQ2000 */
	void logger_cb(LogEvent*);
	void connected_cb(ConnectedEvent* ce);
	void contact_sc_cb(StatusChangeEvent* ev);
	void messageack_cb(MessageEvent*);
	void messaged_cb(MessageEvent*);
	void disconnected_cb(DisconnectedEvent*);
	void socket_cb(SocketEvent* );
	void want_auto_resp_cb(ICQMessageEvent* );


/* async calls */
	void connect();
	void reconnect();
	void disconnect();



	void iq(MyElement*); //handle iq-packets
	void presence(MyElement*); //handle presence-packets

	void message(MyElement*); // handle messages
	void sendText(ContactRef c, MyElement*);
	void sendURL(ContactRef c, MyElement*,xmlnode);
	void sendSMS(ContactRef c, MyElement*);

	void s10n(MyElement*); // handle s10n-packets
	void vcard(MyElement*); // handle vcard-requests
	void handleInfoRequests(unsigned int uin);
	void sendVcard(MyElement *, ContactRef c);

	void sendContactPresence(const unsigned int uin, const string& s="");
	void sendSmsPresence(const string& mobile_no, bool=true);
	void broadcastSmsPresence(bool);

	void authorizeUser(unsigned int uin, bool granted);

	friend class Pinger;
};

struct utils {
	static iconv_t converter_2UTF8;
	static iconv_t converter_unUTF8;

	static Status show2status(const char*);
	static string ui2string(const unsigned int);
	static unsigned int string2ui (const char*);
	static string _2UTF8(const string&);
	static string _unUTF8(const string&);
};

#endif
