/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * --------------------------------------------------------------------------*/

/*
 * Added modifications for ICQv7-t project (icqv7-t.sf.net) by 
 *  - Alexandre N. Safiullin
 *
 * Coryright (c) 2001-2002    Alexandre N. Safiullin <alex_@unislabs.com>
 */

#include "icqtransport.h"

/* #define CONTACT_OFFLINE(c) (c->status == STATUS_OFFLINE || c->status == STATUS_NOT_IN_LIST) */

void it_contact_load_roster(session s)
{
    xmlnode roster, item;
	xmlnode pres;
    jid id;
    UIN_t uin;
    contact c;
	char* to;
	char* invisible;

    id = it_xdb_id(s->p,s->id,s->from->server);
    roster = xdb_get(s->ti->xc,id,NS_ROSTER);
    if (roster == NULL)
        return;

    for (item = xmlnode_get_firstchild(roster); item != NULL; item = xmlnode_get_nextsibling(item))
    {
        uin = it_strtouin(xmlnode_get_attrib(item,"jid"));
        if (uin && uin != s->uin)
        {
            c = it_contact_add(s,uin);
            c->status = STATUS_OFFLINE;
            ++s->cct;

			/* suppose, roster won't be offended for its pool using */
			to=jid_full(it_uin2jid(xmlnode_pool(roster),c->uin, 
								   s->from->server));
			pres=jutil_presnew(JPACKET__SUBSCRIBE,to,NULL);
			xmlnode_put_attrib(pres,"from",jid_full(s->id));

			log_debug(ZONE,"sending roster item %s\n",xmlnode2str(pres));
			it_retransmit(s->ti,pres); /* ask the external part to trace this contact */

			invisible=xmlnode_get_attrib(item,"show");
			if(j_strcmp(invisible,"invisible")==0){
				/* we need to add contact into invisible list ;) */
				pres=jutil_presnew(JPACKET__INVISIBLE,to,NULL);
				xmlnode_put_attrib(pres,"from",jid_full(s->id));
				it_retransmit(s->ti,pres);
			}
				
/* 			if (xmlnode_get_attrib(item,"ask")) */
/*                 c->asking = 1; */
        }
    }

    xmlnode_free(roster);
}

contact it_contact_get(session s, UIN_t uin)
{
    contact c;

    for (c = s->contacts; c != NULL; c = c->next)
        if (c->uin == uin) break;

    return c;
}

contact it_contact_add(session s, UIN_t uin)
{
    contact c;
    pool p;

    p = pool_new();
    c = (contact) pmalloco(p,sizeof(_contact));
    c->p = p;

    c->uin = uin;
    c->status = STATUS_NOT_IN_LIST;
    c->s = s;
 
    c->next = s->contacts;
    s->contacts = c;

    return c;
}

void it_contact_remove(contact c)
{
    session s = c->s;
    contact cur, prev;

    for (prev = NULL, cur = s->contacts; cur != c; prev = cur, cur = cur->next);

    if (prev)
        prev->next = cur->next;
    else
        s->contacts = cur->next;

    pool_free(c->p);
}

/* void it_contact_update_addr(contact c, unsigned long ip, unsigned short port, unsigned long real_ip, icqbyte tcp_flag) */
/* { */
/*     log_debug(ZONE,"%d",(int) tcp_flag); */
/*     memset(&c->addr,0,sizeof(struct sockaddr_in)); */
/*     c->addr.sin_family = AF_INET; */
/*     c->addr.sin_addr.s_addr = ip; */
/*     c->addr.sin_port = port; */
/*     c->addr_real.s_addr = real_ip; */
/*     c->tcp_flag = tcp_flag; */
/*     c->port = (int) port; */
/* } */

/* void it_contact_set_status(contact c, icqstatus status) */
/* { */
/*     int delta = (c->status != status); */

/*     c->status = status; */
/*     if (delta) */
/*         it_contact_send_presence(c); */
/* } */

/* void it_contact_send_presence(contact c) */
/* { */
/*     session s = c->s; */
/*     xmlnode pres, x; */
/*     jid id; */

/*     pres = jutil_presnew(CONTACT_OFFLINE(c) ? JPACKET__UNAVAILABLE : JPACKET__AVAILABLE,jid_full(s->id),NULL); */

/*     id = it_uin2jid(xmlnode_pool(pres),c->uin,s->from->server); */
/*     xmlnode_put_attrib(pres,"from",jid_full(id)); */

/*     switch (c->status) */
/*     { */
/*     case STATUS_AWAY: */
/*         x = xmlnode_insert_tag(pres,"show"); */
/*         xmlnode_insert_cdata(x,"away",-1); */
/*         break; */
/*     case STATUS_DND: */
/*         x = xmlnode_insert_tag(pres,"show"); */
/*         xmlnode_insert_cdata(x,"dnd",-1); */
/*         break; */
/*     case STATUS_NA: */
/*         x = xmlnode_insert_tag(pres,"show"); */
/*         xmlnode_insert_cdata(x,"xa",-1); */
/*         break; */
/*     case STATUS_OCCUPIED: */
/*         x = xmlnode_insert_tag(pres,"show"); */
/*         xmlnode_insert_cdata(x,"dnd",-1); */
/*         break; */
/*     case STATUS_FREE_CHAT: */
/*         x = xmlnode_insert_tag(pres,"show"); */
/*         xmlnode_insert_cdata(x,"chat",-1); */
/*         break; */
/*     default: */
/*         break; */
/*     } */

/*     it_deliver(s->ti,pres); */
/* } */

void it_contact_subscribe(contact c)
{
    session s = c->s;
    iti ti = s->ti;
    xmlnode x, roster;
    jid id;
    pool p;
    char uin[16];

    x = jutil_presnew(JPACKET__SUBSCRIBED,jid_full(s->id),NULL);
    p = xmlnode_pool(x);
    xmlnode_put_attrib(x,"from",jid_full(it_uin2jid(p,c->uin,s->from->server)));

    id = it_xdb_id(p,s->id,s->from->server);
    roster = xdb_get(ti->xc,id,NS_ROSTER);
    if (roster == NULL)
    {
        roster = xmlnode_new_tag("query");
        xmlnode_put_attrib(roster,"xmlns",NS_ROSTER);
    }

    snprintf(uin,15,"%lu",c->uin);

    if (xmlnode_get_tag(roster,spools(p,"item?jid=",uin,p)) == NULL)
    {
        xmlnode_put_attrib(xmlnode_insert_tag(roster,"item"),"jid",uin);
        if (xdb_set(ti->xc,id,NS_ROSTER,roster))
        {
            log_error(ti->i->id,"Roster set failed for '%s'",jid_full(s->id));
        }
    }
    xmlnode_free(roster);

    c->status = STATUS_OFFLINE;
    ++s->cct;

    it_deliver(ti,x); // subscribed notification
}

void it_contact_unsubscribe(contact c)
{
    session s = c->s;
    iti ti = s->ti;
    xmlnode x;
    jid id;
    xmlnode roster, item;
    pool p;
    char uin[16];

/*     x = jutil_presnew(JPACKET__SUBSCRIBED,jid_full(s->id),NULL); */
    x = jutil_presnew(JPACKET__UNSUBSCRIBED,jid_full(s->id),NULL);
    p = xmlnode_pool(x);
    xmlnode_put_attrib(x,"from",jid_full(it_uin2jid(p,c->uin,s->from->server)));

    id = it_xdb_id(p,s->id,s->from->server);
    roster = xdb_get(ti->xc,id,NS_ROSTER);

    if (roster != NULL)
    {
        snprintf(uin,15,"%lu",c->uin);
        item = xmlnode_get_tag(roster,spools(p,"item?jid=",uin,p));

        if (item != NULL)
        {
            xmlnode_hide(item);
            if (xdb_set(ti->xc,id,NS_ROSTER,roster))
            {
                log_error(ti->i->id,"Roster set failed for '%s'",jid_full(s->id));
            }
        }
        xmlnode_free(roster);

        c->status = STATUS_NOT_IN_LIST;
        --s->cct;

/* let's ignore auth requests
        if (c->asking == 0) */
		it_contact_remove(c);
    }

    it_deliver(ti,x);
}

/* void it_contact_ask(void *arg) */
/* { */
/*     xmlnode x = (xmlnode) arg; */
/*     contact c = (contact) xmlnode_get_vattrib(x,"c"); */
/*     session s = c->s; */
/*     iti ti = s->ti; */
/*     pool p = xmlnode_pool(x); */
/*     xmlnode roster, item; */
/*     jid id; */
/*     char uin[16]; */

/*     xmlnode_hide_attrib(x,"c"); */

/*     id = it_xdb_id(p,s->id,s->from->server); */
/*     roster = xdb_get(ti->xc,id,NS_ROSTER); */
/*     if (roster == NULL) */
/*     { */
/*         roster = xmlnode_new_tag("query"); */
/*         xmlnode_put_attrib(roster,"xmlns",NS_ROSTER); */
/*     } */

/*     snprintf(uin,15,"%lu",c->uin); */

/*     item = xmlnode_get_tag(roster,spools(p,"item?jid=",uin,p)); */
/*     if (item == NULL) */
/*     { */
/*         item = xmlnode_insert_tag(roster,"item"); */
/*         xmlnode_put_attrib(item,"jid",uin); */
/*     } */

/*     if (xmlnode_get_attrib(item,"ask") == NULL) */
/*     { */
/*         xmlnode_put_attrib(item,"ask","subscribe"); */
/*         if (xdb_set(ti->xc,id,NS_ROSTER,roster)) */
/*         { */
/*             log_error(ti->i->id,"Roster set failed for '%s'",jid_full(s->id)); */
/*         } */
/*     } */

/*     c->asking = 1; */
/*     xmlnode_free(roster); */
/*     it_deliver(ti,x); */
/* } */

/* void it_contact_subscribed(contact c, jpacket jp) */
/* { */
/*     session s = c->s; */
/*     iti ti = s->ti; */
/*     xmlnode roster, item; */
/*     jid id; */
/*     pool p = jp->p; */
/*     char uin[16]; */

/*     c->asking = 0; */

/*     id = it_xdb_id(p,s->id,s->from->server); */
/*     if ((roster = xdb_get(ti->xc,id,NS_ROSTER)) == NULL) */
/*     { */
/*         if (c->status == STATUS_NOT_IN_LIST) it_contact_remove(c); */
/*         it_message_send(c,it_message_create(jp->x,c->uin,MSGTYPE_AUTHORIZE)); */
/*         return; */
/*     } */

/*     snprintf(uin,15,"%lu",c->uin); */

/*     item = xmlnode_get_tag(roster,spools(p,"item?jid=",uin,p)); */
/*     if (item != NULL && xmlnode_get_attrib(item,"ask")) */
/*     { */
/*         xmlnode_hide_attrib(item,"ask"); */

/*         if (xdb_set(ti->xc,id,NS_ROSTER,roster)) */
/*         { */
/*             log_error(ti->i->id,"Roster set failed for '%s'",jid_full(s->id)); */
/*         } */
/*     } */

/*     xmlnode_free(roster); */
/*     it_message_send(c,it_message_create(jp->x,c->uin,MSGTYPE_AUTHORIZE)); */
/*     if (c->status == STATUS_NOT_IN_LIST) it_contact_remove(c); */
/* } */

void it_contact_free(session s)
{
    contact c;
    pool p;

    c = s->contacts;
    while (c)
    {
        p = c->p;
        c = c->next;
        pool_free(p);
    }
}

/* state: invisible - contact doesn't ses us
          visible   - contact does see us when we're invisible
          <empty>   - regular situation
		  away/xa/etc - contact sees us in the given state (libicq2000 doesn't
		                support it for now
*/
void it_contact_see_as(contact c, const char* state){

	session s = c->s;
    iti ti = s->ti;
    xmlnode roster, item;
    jid id;
    pool p=pool_new();
    char uin[16];

    id = it_xdb_id(p,s->id,s->from->server);
    roster = xdb_get(ti->xc,id,NS_ROSTER);

    if (roster != NULL)
    {
        snprintf(uin,15,"%lu",c->uin);
        item = xmlnode_get_tag(roster,spools(p,"item?jid=",uin,p));

        if (item != NULL)
        {
			if(state) xmlnode_put_attrib(item,"show",state);
			else xmlnode_hide_attrib(item,"show");

            if (xdb_set(ti->xc,id,NS_ROSTER,roster))
            {
                log_error(ti->i->id,"Roster set failed for '%s'",jid_full(s->id));
            }
        }
        xmlnode_free(roster);

	}
	pool_free(p);
}
