/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * icqtransport.c - entry and exit point, loads config, registers callbacks
 * --------------------------------------------------------------------------*/

/*
 * Added modifications for ICQv7-t project (icqv7-t.sf.net) by 
 *  - Alexandre N. Safiullin
 *
 * Coryright (c) 2001      Alexandre N. Safiullin <alex_@unislabs.com>
 */

#include "icqtransport.h"

result it_receive(instance i, dpacket d, void *arg);
void it_shutdown(void *arg);
void it_external(iti ti, jpacket jp);

void icqtrans(instance i, xmlnode x)
{
    iti ti;
    pool p = i->p;
    xmlnode config;

    log_debug(ZONE,"ICQ Transport, initializing for section '%s'",i->id);

    /* create new transport instance */
    ti = pmalloco(p,sizeof(_iti));
    ti->i = i;
    ti->xc = xdb_cache(i);

    config = xdb_get(ti->xc,jid_new(xmlnode_pool(x),"config@-internal"),"jabber:config:icqtrans");
    if (config == NULL)
    {
        log_error(i->id,"Configuration not found!");
        return;
    }

    ti->reg_inst = pstrdup(p,xmlnode_get_tag_data(config,"instructions"));
    if (ti->reg_inst == NULL)
    { /* this disables registration */
        log_debug(i->id,"Registration instructions not found");
    }

    ti->search_inst = pstrdup(p,xmlnode_get_tag_data(config,"search"));
    if (ti->search_inst == NULL)
    {
        log_debug(i->id,"Search instructions not found");
    }

    ti->msg_chat = xmlnode_get_tag(config,"chat") ? 1 : 0;

	ti->external_part=pstrdup(p,xmlnode_get_tag_data(config,"external"));
	ti->sms_id=pstrdup(p,xmlnode_get_tag_data(config,"sms"));

	/* TODO:
	   Maybe to organize exchange of p_ROUTE packets */
/*     ti->dnsrv_id = pstrdup(p,xmlnode_get_tag_data(config,"dnsrv/id")); */
/*     if (ti->dnsrv_id) */
/*     { */
/*         char *host; */
/*         int delay; */

/*         host = xmlnode_get_tag_data(config,"dnsrv/host"); */
/*         ti->dnsrv_host = pstrdup(p,host ? host : "icq.mirabilis.com"); */
/*         delay = j_atoi(xmlnode_get_tag_data(config,"dnsrv/delay"),300);  */
/* 5 minute default */

         /* start dnsrv requests */ 
/*         register_beat(delay,it_dnsresolve,(void *) ti); */
/*         pth_spawn(PTH_ATTR_DEFAULT,it_init_dns,(void *) ti); */
/*     } */

    ti->sessions = xhash_new(j_atoi(xmlnode_get_tag_data(config,"prime"),509));
    ti->vcard = xmlnode_new_tag_pool(p,"vCard");
    xmlnode_put_attrib(ti->vcard,"xmlns",NS_VCARD);
    xmlnode_insert_node(ti->vcard,xmlnode_get_firstchild(xmlnode_get_tag(config,"vCard")));

    ti->admin = xmlnode_dup(xmlnode_get_tag(config,"admin"));
    ti->start = time(NULL);

    register_phandler(i,o_DELIVER,it_receive,(void *) ti);
    register_shutdown(it_shutdown,(void *) ti);

    xmlnode_free(config);
}

void it_retransmit(iti ti, xmlnode x) {
	char* to;
	jpacket jp=jpacket_new(x);
	
	if(jp->to!=NULL){
		jid_set(jp->to,ti->external_part,JID_SERVER);
		to=jid_full(jp->to);
	}
	else to=ti->external_part;

	xmlnode_put_attrib(x,"to",to);
	log_error("ICQv7-t","retransmitting packet: %s",xmlnode2str(x));
	it_deliver(ti,x);
}

/* all Jabber packets are received here */
result it_receive(instance i, dpacket d, void *arg) {
    iti ti = (iti) arg;
    jpacket jp;
    session s;
	
	log_error(ti->i->id,"Packet received: %s\n",xmlnode2str(d->x)); 
	switch(d->type) {
		case p_ROUTE: {
		/* ignore */
		return r_PASS;
    }
			
	case p_NONE:
	case p_NORM:
		jp = jpacket_new(d->x);
		break;

    default:
		return r_ERR;
    }

    if (!jp->from ||/* !jp->from->user ||*/ jp->type == JPACKET_UNKNOWN /* || jpacket_subtype(jp) == JPACKET__ERROR */)
    { /* ignore invalid packets */
        xmlnode_free(jp->x);
        return r_DONE;
    }

    s = (session) xhash_get(ti->sessions,jid_full(jid_user(jp->from)));
    if (s != NULL) {
		log_error("ICQv7-t","Session's found");
        if (s->exit_flag) {
			log_error("ICQv7-t","exit flag enabled");
            if (jp->type != JPACKET_PRESENCE){
                jutil_error(jp->x,TERROR_NOTFOUND);
                it_deliver(ti,jp->x);
			}
            else
				xmlnode_free(jp->x);
		}
        else {
			jp->aux1 = (void *) s;
            mtq_send(s->q,jp->p,it_session_jpacket,(void *) jp);
		}
	}
	else {
		log_error("ICQv7-t","Session's not found, from:%s",jid_full(jp->from));
		/* not into session, maybe from external counterpart? */
		if(j_strcmp(jid_full(jp->from),ti->external_part)==0)
				/* packet from external part */
			it_external(ti,jp);
		else 
			if(jpacket_subtype(jp)!=JPACKET__ERROR)
				it_unknown(ti,jp);
			else
				xmlnode_free(jp->x);
	}
	return r_DONE;
}

// function to handle packets from external part
void it_external(iti ti, jpacket jp){
	session s;
	xmlnode iq, fatal;
	char *ns, *jid;

	switch (jp->type) {
    case JPACKET_IQ: {
		iq=jp->iq;
		ns=xmlnode_get_attrib(iq,"xmlns");
		jid=xmlnode_get_data(xmlnode_get_tag(iq,"jid"));
		s = (session) xhash_get(ti->sessions,jid);

		if(j_strcmp(ns,NS_CONNECTION)==0){
			/* connection related stuff */

			switch(jpacket_subtype(jp)){
			case JPACKET__SET:
				/* reconnect or confirmation */

				if(s!=NULL)
					if(s->confirmed){ /* reconnect */
						
						if(s->presence!=NULL){
							xmlnode_put_attrib(s->presence->x,"from",jid_full(s->presence->to));
							xmlnode_put_attrib(s->presence->x,"to",jid_full(jid_user(s->presence->from)));
							it_deliver(s->ti,s->presence->x);
						}
						else
							log_error("ICQv7-t","Internal error, presence_packet==NULL!");
						it_session_connected(s);
					}
					else{
						it_session_confirmed(s);
						s->confirmed=1;
					}
				
				else
					log_error("ICQv7-t","Internal error! Disconnect message into non-existing session");
					
				break;
			case JPACKET__ERROR:
				/* disconnect occured */
				log_error("ICQv7-t","Disconnect occured");

				fatal=xmlnode_get_tag(iq,"fatal");
				if(s!=NULL&&!s->exit_flag){
					s->connected=0;
					log_debug(ZONE,"ICQv7-t> error into session: %s",s->id);
					it_session_error(s,xmlnode_get_tag(jp->x,"error"),fatal?1:0);
				}
				else 
					log_error("ICQv7-t","Internal error! Disconnect message into non-existing session");
				break;
			default:
				break;
						
			}
		}
	}
	default: break;
	}
	xmlnode_free(jp->x);
	return; 
}
	
/* void *it_init_dns(void *arg) */
/* { */
     /* wait a few seconds to make sure dnsrv starts up */ 
/*     pth_nap(pth_time(2,0)); */
/*     it_dnsresolve(arg); */
/*     return NULL; */
/* } */

/* result it_dnsresolve(void *arg) */
/* { */
/*     iti ti = (iti) arg; */
/*     xmlnode route; */

     /* resolve ICQ server periodically */ 
/*     route = xmlnode_new_tag("route"); */
/*     xmlnode_put_attrib(route,"to",ti->dnsrv_id); */
/*     xmlnode_put_attrib(route,"from",ti->i->id); */
/*     xmlnode_put_attrib(route,"type","request"); */
/*     xmlnode_put_attrib(xmlnode_insert_tag(route,"iq"),"to",ti->dnsrv_host); */
/*     it_deliver(ti,route); */

/*     return r_DONE; */
/* } */

/* void it_sessions_end(xht h, const char *key, void *val, void *arg) */
/* { */
/*     it_session_end((session) val); */
/* } */

void *it_shutdown_wait(void *arg)
{
    iti ti = (iti) arg;

    while (ti->sessions_count)
        pth_nap(pth_time(1,0));

    pth_cond_notify(&ti->sd_cond,FALSE);
    return NULL;
}

void it_shutdown(void *arg)
{
    iti ti = (iti) arg;
    pth_mutex_t mutex;

    log_notice(ZONE,"ICQ Transport, shutting down");

    ti->shutdown = 1;
/*     if (ti->sessions_count) */
/*     { */
/*         pth_mutex_init(&mutex); */
/*         pth_cond_init(&ti->sd_cond); */

/*         xhash_walk(ti->sessions,it_sessions_end,NULL); */

/*         pth_spawn(PTH_ATTR_DEFAULT,it_shutdown_wait,(void *) ti); */

/*         pth_mutex_acquire(&mutex,FALSE,NULL); */
/*         if (ti->sessions_count != 0) */
/*             pth_cond_await(&ti->sd_cond,&mutex,NULL); */
/*         pth_mutex_release(&mutex); */
/*     } */

    xhash_free(ti->sessions);
    ti->sessions = NULL;
}
