/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * --------------------------------------------------------------------------*/

/*
 * Added modifications for ICQv7-t project (icqv7-t.sf.net) by 
 *  - Alexandre N. Safiullin
 *
 * Coryright (c) 2001      Alexandre N. Safiullin <alex_@unis-ru.com>
 */

#include "jabberd.h"
#include "protocol.h"

#include <icqv7t.h>

#define RESEND_DELAY 10
#define MAX_RETRY_ATTEMPTS 8
#define KEEPALIVE 30
#define RECV_SEQ_MAX 16
#define PACKET_DEBUG

typedef struct session_st *session, _session;

/* ICQ Transport Instance */
typedef struct
{
    instance i;
    xdbcache xc;
    xmlnode vcard, admin;
    xht sessions;
    session ka_head, ka_tail, rt_list;
    int sessions_count;
    char *reg_inst;
    char *search_inst;
    char *external_part;
    char *sms_id; // id of sms-t
    int msg_chat;
  /*     char *dnsrv_id; */
  /*     char *dnsrv_host; */
    time_t start;
    int shutdown;
    pth_mutex_t ka_mutex, rt_mutex;
    pth_cond_t sd_cond;
} *iti, _iti;

typedef struct
{
    char *nick;
    char *first;
    char *last;
    char *email;
    char *email2;
    char *old_email;
    char *city;
    char *state;
    char *phone;
    char *fax;
    char *street;
    char *cellphone;
    unsigned long zipcode;
    unsigned short country_code;
    int tz_offset;
    int auth;
    int web_aware;
    int hide_ip;
} meta_gen;

typedef struct
{
    unsigned short age;
    unsigned short gender;
    char *homepage;
    unsigned short year;
    unsigned short month;
    unsigned short day;
    unsigned short languages[3];
} meta_more;

typedef struct
{
    char *city;
    char *state;
    char *phone;
    char *fax;
    char *street;
    unsigned long zipcode;
    unsigned short country_code;
    char *company;
    char *dept;
    char *title;
    unsigned short job;
    char *homepage;
} meta_work;

typedef void (*meta_info_cb)(session s, unsigned short type, void *data, void *arg);
typedef void (*meta_search_cb)(session s, UIN_t uin, meta_gen *res, void *arg);

typedef struct
{
    pool p;
    void *cb;
    void *arg;
} *pendmeta, _pendmeta;

typedef enum
{
    STATUS_NOT_IN_LIST,
    STATUS_OFFLINE,
    STATUS_ONLINE,
    STATUS_AWAY,
    STATUS_DND,
    STATUS_NA,
    STATUS_OCCUPIED,
    STATUS_FREE_CHAT,
    STATUS_INVISIBLE
} icqstatus;

typedef struct contact_st
{
    pool p;
    session s;
    UIN_t uin;
    icqstatus status;
    pth_msgport_t queue;
    int connected;
    int asking;
    struct contact_st *next;
} *contact, _contact;

typedef struct
{
    pool p;
    icqbyte *data;
    unsigned long len, offset;
} *icq_packet, _icq_packet;

typedef struct pqueue_st
{
    icq_packet pak;
    unsigned short seq1;
    unsigned short cmd;
    time_t exp_time;
    int attempts;
    struct pqueue_st *next;
} *pqueue, _pqueue;

typedef enum {stype_normal, stype_register} stype;

struct session_st
{
    pool p;
    jid id;
    jid from;
    mtq q;
    iti ti;
    stype type;
	UIN_t uin;
    ppdb p_db;
    pth_msgport_t queue;
    icqstatus status;
    int connected;
	int confirmed;
    int exit_flag;
    contact contacts;
    int cct;
	jpacket presence;
	unsigned long last_invisible_time;
	int invisible; // whether we're invisible
};

typedef struct
{
    pth_message_t head;
    jpacket jp;
} *jpq, _jpq;

typedef struct
{
    pth_message_t head;
    pool p;
    xmlnode x;
    UIN_t to;
    unsigned short type;
    char **parts;
    int size;
} *imessage, _imessage;

void it_deliver(iti ti, xmlnode x);
session it_session_create(iti ti, jpacket jp);
void it_session_start(void *arg);
void it_session_jpacket(void *arg);
void it_session_connected(void *arg);
void it_session_end(session s);
void it_session_error(session s, xmlnode e, int fatal);

void it_unknown(iti ti, jpacket jp);
void it_unknown_bounce(void *arg);
void it_message(session s, jpacket jp);
void it_vcard(session s, jpacket jp);;
void it_iq(session s, jpacket jp);
void it_iq_vcard_server(iti ti, jpacket jp);
void it_iq_browse_server(iti ti, jpacket jp);
void it_iq_time(iti ti, jpacket jp);
void it_iq_version(iti ti, jpacket jp);
void it_iq_last(iti ti, jpacket jp);
void it_iq_admin(iti ti, jpacket jp);
void it_presence(session s, jpacket jp);
void it_s10n(session s, jpacket jp);
void it_s10n_go(session s, jpacket jp, UIN_t uin);
void it_message_go(session s, jpacket jp, UIN_t uin);

void it_contact_load_roster(session s);
contact it_contact_get(session s, UIN_t uin);
contact it_contact_add(session s, UIN_t uin);
void it_contact_remove(contact c);
void it_contact_update_addr(contact c, unsigned long ip, unsigned short port, unsigned long real_ip, icqbyte force);
void it_contact_set_status(contact c, icqstatus status);
void it_contact_send_presence(contact c);
void it_contact_subscribe(contact c);
void it_contact_unsubscribe(contact c);
void it_contact_ask(void *arg);
void it_contact_subscribed(contact c, jpacket jp);
void it_contact_flush(contact c);
void it_contact_free(session s);

pendmeta it_meta_new(session s, void *cb, void *arg);
void it_meta_packet(session s, icq_decode *packet);

result it_server_ka(void *arg);
result it_server_rt(void *arg);
int it_server_connect(session s);
void it_server_write(session s, icq_packet pak, unsigned short cmd, unsigned short seq1);
void it_server_retransmit(session s);
void it_server_acked_packet(session s, unsigned short seq1);
void it_server_rt_remove(session s);
void it_server_process_packet(session s, icq_decode *packet);

void it_packet_login(session s);
void it_packet_login_1(session s);
void it_packet_register(session s);
void it_packet_ack_messages(session s);
void it_packet_text_code(session s, char *code);
void it_packet_keepalive(session s);
void it_packet_contact_list(session s);
void it_packet_set_status(session s);
void it_packet_reg_info(session s);
void it_packet_meta_get_info(session s, UIN_t uin);
void it_packet_meta_set_gen(session s, meta_gen *info);
void it_packet_meta_set_passwd(session s, char *passwd);
void it_packet_search_uin(session s, UIN_t uin);
void it_packet_search_email(session s, char *email);
void it_packet_search_names(session s, char *nick, char *first, char *last);
void it_packet_message(session s, imessage m);

int it_peer_listen(session s);
void it_peer_connect(mio m, int state, void *arg, icqbyte *buffer, icqshort size);
void it_peer_close(contact c);
void it_peer_send_message(contact c, imessage m);

void it_tcp_parse(mio m, const void *buf, size_t bufsz);
void it_tcp_write(mio m, icq_packet pak);
icq_packet it_tcp_packet_init(session s, icqlong port);
icq_packet it_tcp_packet_ack(session s, icqshort type, icqlong seq, icqbyte *response);
icq_packet it_tcp_packet_msg(session s, imessage m);

void it_icq_message(session s, UIN_t uin, char *ts, unsigned short type, icqbyte *data, unsigned short len);
imessage it_message_create(xmlnode x, UIN_t uin, unsigned short type, ...);
void it_message_send(contact c, imessage);
void it_message_send_server(contact c, imessage m);
int it_message_size(imessage m);
icqbyte *it_message_format(imessage m, icqbyte *buffer);
int it_message_truncate(imessage m);

char *it_convert_utf82windows(pool p, const char *utf8_str);
char *it_convert_windows2utf8(pool p, const char *windows_str);
int it_get_utf82windows_len(const char *utf8_str);

void it_check_endian(void);
jid it_xdb_id(pool p, jid id, char *server);
UIN_t it_strtouin(char *uin);
jid it_uin2jid(pool p, UIN_t uin, char *host);
icqstatus it_show2status(const char *show);
char *it_strrepl(pool p, const char *orig, const char *find, const char *replace);
void put_icqshort(void *buffer, unsigned offset, unsigned short value);
unsigned short get_icqshort(const void *buffer, unsigned offset);
unsigned long get_icqlong(const void *buffer, unsigned offset);
void put_icqlong(void *buffer, unsigned offset, unsigned long value);
unsigned long it_status_to_bits(icqstatus status);
icqstatus it_status_from_bits(unsigned long bits);
const char *it_country_from_code(unsigned short code);
icqshort to_icqshort(unsigned short arg);
icqlong to_icqlong(unsigned long arg);
unsigned long from_icqlong(icqlong arg);
unsigned short from_icqshort(icqshort arg);
unsigned short it_status_to_tcp(icqstatus status);
unsigned short it_format_tcp_msgcommand(unsigned short cmd, icqstatus status);
unsigned long it_checkvalue5(const icqbyte *packet,unsigned packet_len);
void it_scramble(void *buffer, unsigned len, unsigned long checkvalue);
int it_test_v5_server(session s, void *buffer, unsigned len);
void it_delay(xmlnode x, char *ts);
void it_pack_string(icq_packet pak, int len, char *val);
int it_reg_set(session s, xmlnode reg);

#define it_pack_long(pak,val) put_icqlong(pak->data,pak->offset,val); \
    pak->offset += 4;

#define it_pack_short(pak,val) put_icqshort(pak->data,pak->offset,val); \
    pak->offset += 2;

#define it_pack_byte(pak,val)  pak->data[pak->offset++] = val;

#ifdef PACKET_DEBUG
void it_debug_dump(char *zone, void *buffer, unsigned nbytes);
#endif

#define it_jid2uin(x) it_strtouin(x->user)
#define it_deliver(ti, x) deliver(dpacket_new(x),ti->i)
#define ICQ_TCP_MIO mio_handlers_new(NULL,NULL,it_tcp_parse)
#define MIO_ICQ 6
typedef void (*mio_icq_cb)(mio m, int state, void *arg, icq_packet pak);
