/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * iq.c - handles all <iq/> requests for a session
 * --------------------------------------------------------------------------*/

/*
 * Added modifications for ICQv7-t project (icqv7-t.sf.net) by 
 *  - Alexandre N. Safiullin
 *
 * Copyright (c) 2001-2002   Alexandre N. Safiullin <alex_@unislabs.com>
 */

#include "icqtransport.h"
#include <sys/utsname.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

void it_iq_reg_get(session s, jpacket jp);
void it_iq_reg_update(session s, jpacket jp);
void it_iq_reg_remove(session s, jpacket jp);
void it_iq_search_get(session s, jpacket jp);
void it_iq_search_set(session s, jpacket jp);
void it_iq_gateway_get(session s, jpacket jp);
void it_iq_gateway_set(session s, jpacket jp);
void it_iq_browse_user(session s, jpacket jp);

void it_iq(session s, jpacket jp)
{
    char *ns;

/*     if (s->connected == 0)  */
/* 	{ */
/*         jpq jq; */

/*         jq = pmalloco(jp->p,sizeof(_jpq)); */
/*         jq->jp = jp; */
/*         pth_msgport_put(s->queue,(pth_message_t *)jq); */
/*         return; */
/*     } */

    ns = xmlnode_get_attrib(jp->iq,"xmlns");
    switch (jpacket_subtype(jp))
    {
    case JPACKET__GET:
        if (j_strcmp(ns,NS_REGISTER) == 0)
            it_iq_reg_get(s,jp);
        else if (j_strcmp(ns,NS_SEARCH) == 0)
            it_retransmit(s->ti,jp->x);
        else if (j_strcmp(ns,NS_VERSION) == 0)
            it_iq_version(s->ti,jp);
        else if (j_strcmp(ns,NS_TIME) == 0)
            it_iq_time(s->ti,jp);
        else if (j_strcmp(ns,NS_VCARD) == 0)
            jp->to->user ? it_retransmit(s->ti,jp->x) : 
				it_iq_vcard_server(s->ti,jp);
        else if (j_strcmp(ns,NS_LAST) == 0)
            jp->to->user == NULL ? it_iq_last(s->ti,jp) : it_retransmit(s->ti,jp->x);
        else if (j_strcmp(ns,NS_BROWSE) == 0)
            jp->to->user ? it_iq_browse_user(s,jp) : it_iq_browse_server(s->ti,jp);
        else if (j_strcmp(ns,NS_GATEWAY) == 0)
            it_iq_gateway_get(s,jp);
        else if (j_strcmp(ns,NS_ADMIN) == 0)
            it_iq_admin(s->ti,jp);
        else
        {
            jutil_error(jp->x,TERROR_NOTIMPL);
            it_deliver(s->ti,jp->x);
        }
        break;

    case JPACKET__SET:
        if (j_strcmp(ns,NS_REGISTER) == 0)
			if(xmlnode_get_tag(jp->iq,"remove"))
				it_iq_reg_remove(s,jp);
			else {
				jutil_error(jp->x,TERROR_NOTIMPL);  
				it_deliver(s->ti,jp->x);
			}
/*    ?  : it_iq_reg_update(s,jp); */

		else if (j_strcmp(ns,NS_SEARCH) == 0)
            it_retransmit(s->ti,jp->x);
        else if (j_strcmp(ns,NS_GATEWAY) == 0)
            it_iq_gateway_set(s,jp);
        else
        {
            if (j_strcmp(ns,NS_VERSION) == 0 || j_strcmp(ns,NS_TIME) == 0)
                jutil_error(jp->x,TERROR_NOTALLOWED);
            else
                jutil_error(jp->x,TERROR_NOTIMPL);  
            it_deliver(s->ti,jp->x);
        }
        break;

    default:
        xmlnode_free(jp->x);
        break;
    }
}

void it_iq_reg_get(session s, jpacket jp)
{
    iti ti = s->ti;
    xmlnode q, reg, x;

    reg = xdb_get(ti->xc,it_xdb_id(xmlnode_pool(jp->x),s->id,s->from->server),NS_REGISTER);
    if (reg == NULL)
    {
        jutil_error(jp->x,TERROR_NOTFOUND);
        it_deliver(ti,jp->x);
        return;
    }

    jutil_iqresult(jp->x);
    q = xmlnode_insert_tag(jp->x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_REGISTER);
    xmlnode_insert_node(q,xmlnode_get_firstchild(reg));
    xmlnode_free(reg);

    if (xmlnode_get_tag(q,"nick") == NULL)
        xmlnode_insert_tag(q,"nick");

    if (xmlnode_get_tag(q,"first") == NULL)
        xmlnode_insert_tag(q,"first");

    if (xmlnode_get_tag(q,"last") == NULL)
        xmlnode_insert_tag(q,"last");

    if (xmlnode_get_tag(q,"email") == NULL)
        xmlnode_insert_tag(q,"email");

    xmlnode_hide(xmlnode_get_tag(q,"password"));
    xmlnode_insert_tag(q,"password");

    while ((x = xmlnode_get_tag(q,"key")) != NULL) xmlnode_hide(x);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"key"),jutil_regkey(NULL,jid_full(jp->from)),-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"instructions"),ti->reg_inst,-1);
    xmlnode_insert_tag(q,"registered");

    it_deliver(ti,jp->x);
}

void it_iq_reg_remove(session s, jpacket jp)
{
    iti ti = s->ti;
    jid id;
    contact c;
    xmlnode x, pres;

    log_debug(ZONE,"Unregistering user '%s'",jid_full(s->id));

    id = it_xdb_id(jp->p,s->id,s->from->server);
    if (xdb_set(ti->xc,id,NS_REGISTER,NULL) || xdb_set(ti->xc,id,NS_ROSTER,NULL))
    {
        jutil_error(jp->x,(terror){500,"XDB troubles"});
        it_deliver(ti,jp->x);
        return;
    }

    x = jutil_presnew(JPACKET__UNSUBSCRIBE,jid_full(s->id),NULL);
    for (c = s->contacts; c != NULL; c = c->next)
    {
        pres = xmlnode_dup(x);
        xmlnode_put_attrib(pres,"from",jid_full(it_uin2jid(c->p,c->uin,s->from->server)));
        it_deliver(ti,pres);
    }
	xmlnode_put_attrib(x,"from",jid_full(s->from));
    it_deliver(s->ti,x);

    jutil_iqresult(jp->x);
    it_deliver(s->ti,jp->x);

    it_session_end(s);
}

/* void it_iq_reg_result_set(void *arg) */
/* { */
/*     jpacket jp = (jpacket) arg; */
/*     session s = (session) jp->aux1; */

/*     if (it_reg_set(s,jp->iq) == 0) */
/*         jutil_iqresult(jp->x); */
/*     else */
/*         jutil_error(jp->x,TERROR_UNAVAIL); */

/*     it_deliver(s->ti,jp->x); */
/* } */

/* void it_iq_reg_result(session s, unsigned short type, void *data, void *arg) */
/* { */
/*     jpacket jp = (jpacket) arg; */
/*     meta_gen *gen = (meta_gen *) data; */
/*     char *nick, *first, *last, *email; */

/*     if (type == 0) */
/*     { */
/*         jutil_error(jp->x,TERROR_EXTERNAL); */
/*         it_deliver(s->ti,jp->x); */
/*         return; */
/*     } */
/*     else if (type != META_USER_INFO) */
/*         return; */

/*     nick = xmlnode_get_tag_data(jp->iq,"nick"); */
/*     first = xmlnode_get_tag_data(jp->iq,"first"); */
/*     last = xmlnode_get_tag_data(jp->iq,"last"); */
/*     email = xmlnode_get_tag_data(jp->iq,"email"); */

/*     gen->nick = it_convert_utf82windows(jp->p,nick); */
/*     gen->first = it_convert_utf82windows(jp->p,first); */
/*     gen->last = it_convert_utf82windows(jp->p,last); */
/*     gen->email = it_convert_utf82windows(jp->p,email); */

/*     if ((nick != NULL && gen->nick == NULL) || */
/*         (first != NULL && gen->first == NULL) || */
/*         (last != NULL && gen->last == NULL) ||  */
/*         (email != NULL && gen->email == NULL)) */
/*     { */
/*         jutil_error(jp->x,(terror) {406,"Failed to convert UTF-8"}); */
/*         it_deliver(s->ti,jp->x); */
/*     } */
/*     else */
/*     { */
/*         it_packet_meta_set_gen(s,gen); */
/*         jp->aux1 = (void *) s; */
/*         mtq_send(s->q,jp->p,&it_iq_reg_result_set,(void *) jp); */
/*     } */
/* } */

/* void it_iq_reg_update(session s, jpacket jp) */
/* { */
/*     iti ti = s->ti; */
/*     xmlnode q = jp->iq; */
/*     char *passwd; */
/*     int update = 0; */

/*     passwd = it_convert_utf82windows(s->p,xmlnode_get_tag_data(q,"password")); */
/*     if (j_strlen(passwd) > 0 && j_strcmp(s->passwd,passwd)) */
/*     { */
/*         log_debug(ZONE,"Changing '%s' password",jid_full(s->id)); */
         /* Change password */ 
/*         s->passwd = pstrdup(s->p,passwd); */
/*         it_packet_meta_set_passwd(s,passwd); */
/*         update = 1; */
/*     } */
/*     else if (passwd == NULL && xmlnode_get_tag(q,"password")) */
/*     { */
/*         jutil_error(jp->x,(terror) {406,"Failed to convert UTF-8"}); */
/*         it_deliver(ti,jp->x); */
/*         return; */
/*     } */

/*     xmlnode_hide(xmlnode_get_tag(q,"password")); */
/*     xmlnode_insert_cdata(xmlnode_insert_tag(q,"password"),s->passwd,-1); */

/*     if (j_strcmp(s->nick,xmlnode_get_tag_data(q,"nick")) || */
/*         j_strcmp(s->first,xmlnode_get_tag_data(q,"first")) || */
/*         j_strcmp(s->last,xmlnode_get_tag_data(q,"last")) || */
/*         j_strcmp(s->email,xmlnode_get_tag_data(q,"email"))) { */

/*         log_debug(ZONE,"Updating user profile"); */

/*         s->pend_info = it_meta_new(s,(void *) &it_iq_reg_result,(void *) jp); */
/*         it_packet_meta_get_info(s,s->uin); */
/*         return; */
/*     } */

/*     if (update) */
/*     { */
/*         if (it_reg_set(s,jp->iq)) */
/*             jutil_error(jp->x,TERROR_UNAVAIL); */
/*         else */
/*             jutil_iqresult(jp->x); */
/*     } */
/*     else */
/*         jutil_iqresult(jp->x); */

/*     it_deliver(ti,jp->x); */
/* } */

/* void it_iq_search_get(session s, jpacket jp) */
/* { */
/*     iti ti = s->ti; */
/*     xmlnode q; */

/*     if (ti->search_inst == NULL || jp->to->user != NULL) */
/*     { */
/*         jutil_error(jp->x,TERROR_NOTALLOWED); */
/*         it_deliver(s->ti,jp->x); */
/*         return; */
/*     } */

/*     jutil_iqresult(jp->x); */
/*     q = xmlnode_insert_tag(jp->x,"query"); */
/*     xmlnode_put_attrib(q,"xmlns",NS_SEARCH); */

    /* Fill in the values on which we can search. */
/*     xmlnode_insert_tag(q,"username"); */
/*     xmlnode_insert_tag(q,"email"); */
/*     xmlnode_insert_tag(q,"nick"); */
/*     xmlnode_insert_tag(q,"first"); */
/*     xmlnode_insert_tag(q,"last"); */

/*     xmlnode_insert_cdata(xmlnode_insert_tag(q,"instructions"),ti->search_inst,-1); */
/*     xmlnode_insert_cdata(xmlnode_insert_tag(q,"key"),jutil_regkey(NULL,jid_full(jp->from)),-1); */

/*     it_deliver(ti,jp->x); */
/* } */

/* void it_iq_search_result(session s, UIN_t uin, meta_gen *info, void *arg) */
/* { */
/*     jpacket jp = (jpacket) arg; */

/*     if (info != NULL) */
/*     { */
/*         xmlnode item; */
/*         pool p = jp->p; */

/*         log_debug(ZONE,"Search result: %lu",uin); */

/*         item = xmlnode_insert_tag(jp->iq,"item"); */

/*         xmlnode_put_attrib(item,"jid",jid_full(it_uin2jid(p,uin,s->from->server))); */

/*         xmlnode_insert_cdata(xmlnode_insert_tag(item,"email"), */
/*                              it_convert_windows2utf8(p,info->email),-1); */

/*         xmlnode_insert_cdata(xmlnode_insert_tag(item,"nick"), */
/*                              it_convert_windows2utf8(p,info->nick),-1); */

/*         xmlnode_insert_cdata(xmlnode_insert_tag(item,"first"), */
/*                              it_convert_windows2utf8(p,info->first),-1); */

/*         xmlnode_insert_cdata(xmlnode_insert_tag(item,"last"), */
/*                              it_convert_windows2utf8(p,info->last),-1); */
/*     } */
/*     else */
/*         it_deliver(s->ti,jp->x); */
/* } */

/* void it_iq_search_set(session s, jpacket jp) */
/* { */
/*     xmlnode query = jp->iq; */
/*     pool p; */
/*     UIN_t uin; */
/*     char *first, *last, *nick, *email; */

/*     if (s->ti->search_inst == NULL || s->pend_search != NULL || jp->to->user != NULL) */
/*     { */
/*         jutil_error(jp->x,TERROR_NOTALLOWED); */
/*         it_deliver(s->ti,jp->x); */
/*         return; */
/*     } */

/*     p = jp->p; */
/*     uin = it_strtouin(xmlnode_get_tag_data(query,"username")); */
/*     nick = it_convert_utf82windows(p,xmlnode_get_tag_data(query,"nick")); */
/*     first = it_convert_utf82windows(p,xmlnode_get_tag_data(query,"first")); */
/*     last = it_convert_utf82windows(p,xmlnode_get_tag_data(query,"last")); */
/*     email = it_convert_utf82windows(p,xmlnode_get_tag_data(query,"email")); */

    /* Make sure at least ONE of the possible search parameters is specified */
/*     if (uin == 0 && nick == NULL && first == NULL && last == NULL && email == NULL) */
/*     { */
/*         jutil_error(jp->x,(terror) {406,"No valid search parameters specified"}); */
/*         it_deliver(s->ti,jp->x); */
/*         return; */
/*     } */

/*     jutil_iqresult(jp->x); */
/*     xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_SEARCH); */
/*     jpacket_reset(jp); */

/*     s->pend_search = it_meta_new(s,(void *) &it_iq_search_result,(void *) jp); */

    /* Determine the search type. */
/*     if (uin != 0) */
/*         it_packet_search_uin(s,uin); */
/*     else if (email) */
/*         it_packet_search_email(s,email); */
/*     else */
/*         it_packet_search_names(s,nick ? nick : "",first ? first : "",last ? last : ""); */
/* } */

void it_iq_gateway_get(session s, jpacket jp)
{
    if (jp->to->user == NULL)
    {
        xmlnode q;

        jutil_iqresult(jp->x);

        q = xmlnode_insert_tag(jp->x,"query");
        xmlnode_put_attrib(q,"xmlns",NS_GATEWAY);
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"desc"),"Enter the user's UIN",-1);
        xmlnode_insert_tag(q,"prompt");
    }
    else
        jutil_error(jp->x,TERROR_NOTALLOWED);

    it_deliver(s->ti,jp->x);
}

void it_iq_gateway_set(session s, jpacket jp)
{
    char *user, *id;

    user = xmlnode_get_tag_data(jp->iq,"prompt");
    id = user ? spools(jp->p,user,"@",jp->to->server,jp->p) : NULL;
    if (id && it_strtouin(user))
    {
        xmlnode q;

        jutil_iqresult(jp->x);

        q = xmlnode_insert_tag(jp->x,"query");
        xmlnode_put_attrib(q,"xmlns",NS_GATEWAY);
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"prompt"),id,-1);
    }
    else
        jutil_error(jp->x,TERROR_BAD);

    it_deliver(s->ti,jp->x);
}

void it_iq_browse_server(iti ti, jpacket jp)
{
    xmlnode q;

    q = xmlnode_insert_tag(jutil_iqresult(jp->x),"service");
    xmlnode_put_attrib(q,"xmlns",NS_BROWSE);

    xmlnode_put_attrib(q,"type","icq");
    xmlnode_put_attrib(q,"jid",jp->to->server);
    xmlnode_put_attrib(q,"name",xmlnode_get_tag_data(ti->vcard,"FN"));

/*    xmlnode_insert_cdata(xmlnode_insert_tag(q,"ns"),NS_REGISTER,-1); */
/*    xmlnode_insert_cdata(xmlnode_insert_tag(q,"ns"),NS_SEARCH,-1); */
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"ns"),NS_GATEWAY,-1);

    it_deliver(ti,jp->x);
}

void it_iq_browse_user(session s, jpacket jp)
{
    xmlnode browse;

    if (s->type == stype_register)
    {
        jpq jq;

        jq = pmalloco(jp->p,sizeof(_jpq));
        jq->jp = jp;
        pth_msgport_put(s->queue,(pth_message_t *)jq);
        return;
    }

    if (it_jid2uin(jp->from) == 0)
    {
        jutil_error(jp->x,TERROR_BAD);
        it_deliver(s->ti,jp->x);
        return;
    }

    jutil_iqresult(jp->x);
    browse = xmlnode_insert_tag(jp->x,"user");
    xmlnode_put_attrib(browse,"xmlns",NS_BROWSE);
    xmlnode_put_attrib(browse,"jid",jid_full(jid_user(jp->to)));
    xmlnode_put_attrib(browse,"type","user");

    it_deliver(s->ti,jp->x);
}

void it_iq_version(iti ti, jpacket jp)
{
    xmlnode os, x, q;
    struct utsname un;

    x = jutil_iqresult(jp->x);

    q = xmlnode_insert_tag(x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_VERSION);

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"name"),"ICQv7 Transport",-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"version"),VERSION,-1);

    uname(&un);
    os = xmlnode_insert_tag(q,"os");
    xmlnode_insert_cdata(os,un.sysname,-1);
    xmlnode_insert_cdata(os," ",1);
    xmlnode_insert_cdata(os,un.release,-1);

    it_deliver(ti,x);
}

void it_iq_time(iti ti, jpacket jp)
{
    xmlnode x, q;

    x = jutil_iqresult(jp->x);

    q = xmlnode_insert_tag(x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_TIME);

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"utc"),jutil_timestamp(),-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"tz"),tzname[0],-1);

    it_deliver(ti,x);
}

void it_iq_last(iti ti, jpacket jp)
{
    xmlnode x, q;
    char str[10];

    x = jutil_iqresult(jp->x);

    snprintf(str,10,"%d",time(NULL) - ti->start);
    q = xmlnode_insert_tag(x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_LAST);
    xmlnode_put_attrib(q,"seconds",str);

    it_deliver(ti,x);
}

void it_iq_vcard_server(iti ti, jpacket jp)
{
    /* send back transport vCard */
    xmlnode_insert_node(jutil_iqresult(jp->x),ti->vcard);
    it_deliver(ti,jp->x);
}

void it_iq_admin_who(xht h, const char *key, void *data, void *arg)
{
    session s = (session) data;
    xmlnode who = (xmlnode)arg;
    xmlnode x;

    x = xmlnode_insert_tag(who,"presence");
    xmlnode_put_attrib(x,"from",jid_full(s->id));
}

void it_iq_admin(iti ti, jpacket jp)
{
    xmlnode x;

    if (ti->admin && xmlnode_get_tag(ti->admin,spools(jp->p,"read=",jid_full(jid_user(jp->from)),jp->p)) != NULL)
    {
        if((x = xmlnode_get_tag(jp->iq,"who")))
            xhash_walk(ti->sessions,it_iq_admin_who,(void *) x);

        jutil_tofrom(jp->x);
        xmlnode_put_attrib(jp->x,"type","result");
    }
    else
        jutil_error(jp->x,TERROR_NOTALLOWED);

    it_deliver(ti,jp->x);
}
