/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * presence.c - handles presence and s10n from sessions
 * --------------------------------------------------------------------------*/

/*
 * Added modifications for ICQv7-t project (icqv7-t.sf.net) by 
 *  - Alexandre N. Safiullin
 *
 * Copyright (c) 2001-2002   Alexandre N. Safiullin <alex_@unislabs.com>
 */

#include "icqtransport.h"

void it_presence(session s, jpacket jp)
{
	contact c;
	UIN_t uin;
    log_debug(ZONE,"Session[%s], handling presence",jid_full(s->id));

	if(jp->to->user) {
		uin = it_jid2uin(jp->to);
		if (uin == 0 || uin == s->uin)
		{
			jutil_error(jp->x,TERROR_BAD);
			it_deliver(s->ti,jp->x);
			return;
		}
	}

    switch (jpacket_subtype(jp))
    {
    case JPACKET__PROBE:
        if (jp->to->user){
			it_retransmit(s->ti,jp->x);
			return;
		}
        break;

    case JPACKET__AVAILABLE:
    {
        icqstatus status;
        char *text;

        if (jp->to->user){
			c=it_contact_get(s,uin);
			if(c&&!s->invisible)it_contact_see_as(c,NULL);
			it_retransmit(s->ti,jp->x);
			return;
		}

        /* update users presence */
        s->p_db = ppdb_insert(s->p_db,jp->from,jp->x);

		s->invisible=0;

        text = xmlnode_get_tag_data(jp->x,"status");

        status = it_show2status(xmlnode_get_tag_data(jp->x,"show"));
        if (status != s->status)
            s->status = status;

		it_retransmit(s->ti,xmlnode_dup(jp->x)); // notify the external part

		if(s->connected){
			/* echo.... */
			xmlnode_put_attrib(jp->x,"from",jid_full(jp->to));
			xmlnode_put_attrib(jp->x,"to",jid_full(jid_user(jp->from)));
			
			it_deliver(s->ti,jp->x);
		}
		else 
			s->presence=jp;
		/* will be echoed upon confirmation */

		return;
    }

    case JPACKET__UNAVAILABLE:
        if (jp->to->user == NULL)
        { /* kill the session if there are no available resources left */
			
			if(time(NULL)-s->last_invisible_time<=2){
#ifdef DEBUG
				log_error(ZONE,"igroning unavailable, it seems related to invisible");
#endif
				break;
			}

            s->p_db = ppdb_insert(s->p_db,jp->from,jp->x);
            if (ppdb_primary(s->p_db,s->id) == NULL){
							// let's notify the external part 
//				it_retransmit(s->ti,xmlnode_dup(jp->x));
                it_session_end(s);
			}
				
        }
		break;
	case JPACKET__INVISIBLE:
		if(jp->to->user) {
			uin = it_jid2uin(jp->to);
			if (uin == 0 || uin == s->uin)
			{
				jutil_error(jp->x,TERROR_BAD);
				it_deliver(s->ti,jp->x);
				return;
			}
			c=it_contact_get(s,uin);
			if(c&&!s->invisible)it_contact_see_as(c,"invisible");
			it_retransmit(s->ti,jp->x);
			return;
		}
		else {
			s->last_invisible_time=time(NULL);
			s->invisible=1;
			it_retransmit(s->ti,jp->x);
			return;
		}
			
    default:
        break;
    }

    xmlnode_free(jp->x);
}

void it_s10n(session s, jpacket jp)
{
	UIN_t uin;
    pool p = jp->p;

    if (jp->to->user == NULL)
    {
		/* ignore s10n to the transport */
        xmlnode_free(jp->x);
        return;
    }

    uin = it_jid2uin(jp->to);
	if (uin == 0 || uin == s->uin)
	{
        jutil_error(jp->x,TERROR_BAD);
        it_deliver(s->ti,jp->x);
        return;
	}

    if (s->connected == 0)
    {
        jpq jq;

        jq = pmalloco(p,sizeof(_jpq));
        jq->jp = jp;
        pth_msgport_put(s->queue,(pth_message_t *)jq);
    }
    else
        it_s10n_go(s,jp,uin);
}

void it_s10n_go(session s, jpacket jp, UIN_t uin)
{
	contact c;

    if (uin == 0)
        uin = it_jid2uin(jp->to);

    c = it_contact_get(s,uin);

	it_retransmit(s->ti,xmlnode_dup(jp->x)); 

	switch (jpacket_subtype(jp))
	{
    case JPACKET__SUBSCRIBE:
        if (c == NULL)
            c = it_contact_add(s,uin);

        if (c->status == STATUS_NOT_IN_LIST){
            it_contact_subscribe(c); /* add to transport-specific roster */
		}
		else 
			log_error("debug","it_contact_send_presence isn't implemented");
/*             it_contact_send_presence(c); */

/*         jutil_tofrom(jp->x); */
/*         xmlnode_put_attrib(jp->x,"type","subscribed"); */
/*         it_deliver(s->ti,jp->x); */
		xmlnode_free(jp->x);
        break;

    case JPACKET__SUBSCRIBED:
		/* external part deals with it */
		xmlnode_free(jp->x);
        break;

    case JPACKET__UNSUBSCRIBE:
        if (c && c->status != STATUS_NOT_IN_LIST)
            it_contact_unsubscribe(c);

        xmlnode_free(jp->x);
        break;

    case JPACKET__UNSUBSCRIBED:
		/* external part deals with it */
		xmlnode_free(jp->x);
        break;

    default:
        xmlnode_free(jp->x);
        break;
	}
}
