/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * session.c - creates and destorys sessions, and some other stuff
 * --------------------------------------------------------------------------*/

/*
 * Added modifications for ICQv7-t project (icqv7-t.sf.net) by 
 *  - Alexandre N. Safiullin
 *
 * Copyright (c) 2001-2002   Alexandre N. Safiullin <alex_@unislabs.com>
 */

#include "icqtransport.h"

session it_session_create(iti ti, jpacket jp)
{
    pool p;
    session s;
    jid id = jp->from;

    log_debug(ZONE,"Creating session for %s",jid_full(id));

    p = pool_new();
    s = pmalloco(p,sizeof(_session));
    s->p = p;
    s->ti = ti;
    s->q = mtq_new(p);

    s->id = jid_new(p,jid_full(jid_user(jp->from)));
    s->from = jid_new(p,jid_full(jp->to));
    jid_set(s->from,"registered",JID_RESOURCE);

    s->queue = pth_msgport_create("it_session_queue");
    s->type = stype_normal;

	xhash_put(ti->sessions,jid_full(s->id),(void *) s);
    ++ti->sessions_count;

	s->confirmed=0;
	s->presence=NULL;
	s->last_invisible_time=0;
	s->invisible=0;
    return s;
}

void it_session_jpacket(void *arg)
{
    jpacket jp = (jpacket) arg;
    session s = (session) jp->aux1;

    if (s->exit_flag&&
		jpacket_subtype(jp)!=JPACKET__PROBE) /* for invisibility, 'cause it
												sends 'probe's */
    {
#ifdef DEBUG
		log_error(ZONE,"exit flag enabled!");
#endif
		if (jp->type != JPACKET_PRESENCE)
		{
            jutil_error(jp->x,TERROR_NOTFOUND);
            it_deliver(s->ti,jp->x);
        }
        else
            xmlnode_free(jp->x);
        return;
    }

    switch (jp->type)
    {
    case JPACKET_IQ:
        it_iq(s,jp);
        break;
    case JPACKET_PRESENCE:
        it_presence(s,jp);
        break;
	case JPACKET_S10N:
		log_error("debug"," wow! s10n caught: %s\n",xmlnode2str(jp->x));
        it_s10n(s,jp);
        break;
    case JPACKET_MESSAGE:
        it_message(s,jp);
        break;
    default:
        xmlnode_free(jp->x);
    }
}

void it_session_register(session s, jpacket jp)
{
    iti ti = s->ti;

    s->type = stype_normal;

	if (it_reg_set(s,jp->iq) == 0)
    {
        xmlnode x;
        char *from;

        from = jid_full(s->from);

        x = jutil_presnew(JPACKET__SUBSCRIBE,jid_full(s->id),NULL);
        xmlnode_put_attrib(x,"from",from);
        it_deliver(ti,x);

        x = jutil_presnew(JPACKET__PROBE,jid_full(s->id),NULL);
        xmlnode_put_attrib(x,"from",from);
        it_deliver(ti,x);

        jutil_iqresult(jp->x);
        it_deliver(ti,jp->x);
    }
    else
    {
        jutil_error(jp->x,(terror){500,"XDB troubles"});

        it_session_end(s);
        it_deliver(ti,jp->x);
    }
}

void it_session_confirmed(session s){
    jpq jq;
    jpacket jp;

    if (s->exit_flag)
        return;

    log_error("it_session_connected","Session[%s] is connected",jid_full(s->id));

    jq = (jpq) pth_msgport_get(s->queue);
    jp = jq->jp;

    if (s->type != stype_normal)
        it_session_register(s,jp);
	else {
		if (jp->type == JPACKET_PRESENCE){ /* all is done already,
											  it's 'connection' scenario */

			if(jpacket_subtype(jp)!=JPACKET__INVISIBLE)
			/* to set proper status */
				it_retransmit(s->ti,xmlnode_dup(jp->x));
				
			else /* it was 'invisible' */
				xmlnode_put_attrib(jp->x,"type","available");
			
			xmlnode_put_attrib(jp->x,"from",jid_full(jp->to));
			xmlnode_put_attrib(jp->x,"to",jid_full(jid_user(jp->from)));
			it_deliver(s->ti,jp->x);
			
		}		
		else 
			log_error("debug","Internal error!");
	}
	it_session_connected(s);
}

void it_session_connected(void *arg)
{
    jpq jq;
    jpacket jp;
    session s = (session) arg;

	if(s->exit_flag)return;

    s->connected = 1;

    /* flush buffer of packets waiting for login to finish */
    while ((jq = (jpq) pth_msgport_get(s->queue)) != NULL)
    {
        jp = jq->jp;

        /* JPACKET_PRESENCE is never buffered */
        switch (jp->type)
        {
        case JPACKET_IQ:
            it_iq(s,jp);
            break;
        case JPACKET_S10N:
            it_retransmit(s->ti,jp->x);
            break;
        case JPACKET_MESSAGE:
			log_error("debug", "unhandled message discovered, ignore it while");
        default:
            xmlnode_free(jp->x);
        }
    }
}


/* shutdown a session, free's resources, etc */
void it_session_exit(void *arg)
{
    session s = (session) arg;
    iti ti = s->ti;
    jpq jq;

    log_debug(ZONE,"Session[%s], exiting",jid_full(s->id));

    xhash_zap(ti->sessions,jid_full(s->id));

    if (s->contacts)
        it_contact_free(s);

    while ((jq = (jpq) pth_msgport_get(s->queue)) != NULL)
    {
        xmlnode x = jq->jp->x;
        jutil_error(x,TERROR_NOTFOUND);
        it_deliver(ti,x);
    }
    pth_msgport_destroy(s->queue);

    pool_free(s->p);

    --ti->sessions_count;
}

void it_session_regerr(session s, xmlnode e)
{
    jpq jq;
    xmlnode x;

    jq = (jpq) pth_msgport_get(s->queue);
    x = jq->jp->x;

	xmlnode_put_attrib(x,"type","error");
	xmlnode_insert_tag_node(x,e);
	jutil_tofrom(x);

    it_deliver(s->ti,x);
}

void it_session_unavail(session s, char *msg)
{
    xmlnode pres;

    pres = jutil_presnew(JPACKET__UNAVAILABLE,jid_full(s->id),NULL);
    xmlnode_put_attrib(pres,"from",jid_full(s->from));
    xmlnode_insert_cdata(xmlnode_insert_tag(pres,"status"),msg,-1);
    it_deliver(s->ti,pres);
}

/* regular exit (i.e. user sent 'unavailable' or asked to remove his registration */
void it_session_end(session s)
{
	xmlnode x;

    if (s->exit_flag == 1)
        return;

	log_debug(ZONE,"Killing session[%s]",jid_full(s->id));

	s->exit_flag=1;

	x=jutil_presnew(JPACKET__UNAVAILABLE,NULL,NULL);
    xmlnode_put_attrib(x,"from",jid_full(s->id));
	it_retransmit(s->ti,x);

	it_session_unavail(s,"Disconnected");
	
	mtq_send(s->q,s->p,it_session_exit,(void *) s);
}


/* scenario:
   1. the external backend discovered troubles
*/

void it_session_error(session s, xmlnode error, int fatal)
{
	xmlnode kill_packet;
    if (s->exit_flag == 1)
        return;
			
	if(fatal||s->type!=stype_normal){ //exit condition

		s->exit_flag=1; /* we're existing */
		if(s->type == stype_normal) /* e.g. password changed outside */
			it_session_unavail(s,xmlnode_get_data(error));
		else { /* e.g. attempt to register with 
				 bad password */
			if(!fatal){
				/* need to kill the external session */
				
				/* let's do it by emulating unavailable presence */
				kill_packet = jutil_presnew(JPACKET__UNAVAILABLE,NULL,NULL);
				xmlnode_put_attrib(kill_packet,"from",jid_full(s->id));
				it_retransmit(s->ti,kill_packet);
			}
			it_session_regerr(s,error);
		}
		mtq_send(s->q,s->p,it_session_exit,(void *) s);
	}
	else { // !fatal && stype_normal
		it_session_unavail(s,strcat(xmlnode_get_data(error)," ICQv7-t will try to reconnect."));
	}
}
