//============================================================================
// Project:       Bedrock - Foundational objects
// Filename:      Application.cpp
// Description:   bedrock::Application implementation
// Created at:    Thu Jul 19 13:15:16 2001
// Modified at:   Mon Jul 30 19:32:10 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
// $Id: Application.cpp,v 1.1 2001/12/16 20:02:26 dr_pepper Exp $
//============================================================================
#include <signal.h>

#include "bedrock.h"
using namespace bedrock;
using namespace std;


Application* Application::_app_instance = NULL;

void Application::signal_callback(int signal)
{
    if (_app_instance->_started)
    {
	_app_instance->_shutdown_lock.lock();
	_app_instance->_shutdown_code   = signal;
	_app_instance->_signal_occurred = true;
	_app_instance->_shutting_down = true;
	_app_instance->_shutdown_signal.signal();
	_app_instance->_shutdown_lock.unlock();
    }
}

void Application::exit(int ec, const string& message)
{
    cerr << "Application exit requested (" << ec << ")" << endl;
    if (!message.empty())
    {
	cout << "Exit message: " << message << endl;
    }
    ::exit(ec);
}

int Application::start()
{
    if (_app_instance == NULL)
	_app_instance = new Application();
    assert(_app_instance->_started == false);

    // Lock the mutex and wait to get signaled
    MutexProxy mp(_app_instance->_shutdown_lock);
    _app_instance->_started = true;
    while (!_app_instance->_shutting_down)
    {
	_app_instance->_shutdown_signal.wait();
    }
    
    _app_instance->_started = false;
    // Process the return code; do special things if a signal occurred.
    if (_app_instance->_signal_occurred)
    {
	cout << "Signal occurred: " << _app_instance->_shutdown_code << endl;
	return 0;
    }
    else
    {
	cout << "Application stop requested (" << _app_instance->_shutdown_code << ")" << endl;
	if (!_app_instance->_shutdown_msg.empty())
	    cout << "Shutdown message: " << _app_instance->_shutdown_msg << endl;
	return _app_instance->_shutdown_code;
    }
}

void Application::stop(int rc, const string& message)
{
    assert(_app_instance != NULL);
    assert(_app_instance->_started == true);

    MutexProxy mp(_app_instance->_shutdown_lock);
    _app_instance->_shutting_down = true;
    _app_instance->_shutdown_code = rc;
    _app_instance->_shutdown_msg  = message;
    _app_instance->_shutdown_signal.signal();
}

void Application::registerShutdownSignal(int signal_id)
{
    if (_app_instance == NULL)
	_app_instance = new Application();
    
    assert(_app_instance->_started == false);
    signal(signal_id, &Application::signal_callback);
}
