//============================================================================
// Project:       Bedrock -- Foundational Objects
// Filename:      bedrock_debug.h
// Description:   bedrock::debug namespace
// Created at:    Thu Jul 19 15:07:42 2001
// Modified at:   Thu Jul 19 15:43:35 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
// $Id: bedrock_debug.h,v 1.1 2001/12/16 20:02:26 dr_pepper Exp $
//============================================================================
#ifndef INCL_BEDROCK_DEBUG_H
#define INCL_BEDROCK_DEBUG_H

#include <iostream>
#include <string>

#ifdef DEBUG
namespace bedrock
{
    namespace debug
    {
        class Tracer
        {
        public:
            Tracer(const char* name, std::ostream& os = std::cerr):
                _name(name), _os(os)
                { _os << "(" << pthread_self() << ") --> " << _name << std::endl; }
            ~Tracer()
                { _os << "(" << pthread_self() << ") <-- " << _name << std::endl; }

            template<class T>
            std::ostream& operator<<(T value)
                { 
                    _os << "(" << pthread_self() << ") -- " << _name << ": "<< value; 
                    return _os; 
                }
        private:
            std::string         _name;
            std::ostream&       _os;
        };
    }

    /* 
       Function to simplify error checking.  The argument passed into the
       function should be return value of a function which returns 0 for
       success and errno for failure.  In particular, pthread_* calls
       behave this way. Errors for these functions is considered fatal, so
       the validating function will abort if the result code is non-zero.
    */
    inline void abortingCall(int funcresult)
	{
	    if (funcresult != 0)
	    {
		std::cerr << "Function call failed with errno = " << funcresult 
			  << "(" << strerror(funcresult) << "). Aborting." << std::endl;
		abort();
	    }
	}

}
#else
namespace bedrock
{
    namespace debug
    {
        class Tracer
        {
        public:
            Tracer(const char* name, std::ostream& os = std::cerr) {}

            template<class T>
            Tracer& operator<<(T value)
                { return *this; }

            Tracer& operator << (std::ostream& f(std::ostream&))
                { return *this; }
        };
    }

    inline void abortingCall(int funcresult)
	{}

}

#endif

#endif



