//============================================================================
// Project:       Bedrock -- Foundational Objects
// Filename:      Address.cpp
// Description:   bedrock::net::Address implementation
// Created at:    Thu Jul 19 10:00:51 2001
// Modified at:   Mon Jul 30 19:27:36 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
// $Id: Address.cpp,v 1.1 2001/12/16 20:02:26 dr_pepper Exp $
//============================================================================
#include "bedrock.h"
using namespace bedrock;
using namespace bedrock::net;
using namespace std;

/**
 * Thread safe version of the function inet_ntoa provided
 * on most operating systems.
 * @param addr Numeric address to convert to a string
 * @param result Reference to a string in which to store
 * the string representation of the numeric address.
 */
void Address::inet_ntoa(unsigned char* addr, string& result)
{
    // Take the numeric address in the hostent* and translate
    // it into human readable format. This is an
    // alternative to using inet_ntoa since inet_ntoa is not 
    // threadsafe on most platforms.
    char resultbuf[16];
    snprintf(resultbuf, 16, "%d.%d.%d.%d", addr[0],addr[1],addr[2],addr[3]); 
    result = resultbuf;
}

/**
 * Thread safe version of the function inet_ntoa provided
 * on most operating systems; alternative which returns a new string.
 * @param addr Numeric address to convert to a string
 * @returns New string representation of numeric address
 */
string Address::inet_ntoa(unsigned char* addr)
{
    // Take the numeric address in the hostent* and translate
    // it into human readable format. This is an
    // alternative to using inet_ntoa since inet_ntoa is not 
    // threadsafe on most platforms.
    char resultbuf[16];
    snprintf(resultbuf, 16, "%d.%d.%d.%d", addr[0],addr[1],addr[2],addr[3]); 
    return string(resultbuf);
}
