//============================================================================
// Project:       Bedrock -- Foundational Objects
// Filename:      KarmaFactory.cpp
// Description:   bedrock::net::KarmaFactory -- implementation
// Created at:    Thu Jul 26 10:00:51 2001
// Modified at:   Mon Jul 30 19:28:28 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
//============================================================================
#include "bedrock.h"
using namespace bedrock;
using namespace bedrock::net;
using namespace std;


KarmaFactory::KarmaFactory(ThreadPool& tpool) : _timer(1)
{
    _timer.schedule(1, &tpool, 0, callback::wrap(this, &KarmaFactory::_resetKarma), -1);
}

/**
 * Create a new karma object
 * @param read_bytes the number of bytes one may read per second
 * @param write_bytes the number of bytes one may write per second
 * @param tpool the threadpool to process callbacks in (on reset)
 * @param tkey the thread key to use when enqueuing callbacks on reset
 */
Karma* KarmaFactory::create(int read_bytes, int write_bytes, ThreadPool& tpool,
			    int tkey)
{
    Karma *kp = new Karma(read_bytes, write_bytes, tpool, tkey, *this);
    MutexProxy mp(_lock);
    _k_list.push_front(kp);
    return kp;
}

/**
 * Start watching this karma object
 */
void KarmaFactory::_register(Karma *karma)
{
    MutexProxy mp(_lock);
    _k_list.push_front(karma);
}

/**
 * Stop watching this karma object
 */
void KarmaFactory::_unregister(Karma *karma)
{
    MutexProxy mp(_lock);
    _k_list.remove(karma);
}

/**
 * for_each callback -- reset this karma object's read/write values
 */
void KarmaFactory::_do_reset(Karma* kp)
{
    kp->_reset();
}

/**
 * timer callback -- another time interval has passed -- tell everyone to
 * reset their time values.
 */
void KarmaFactory::_resetKarma()
{
    MutexProxy mp(_lock);
    for_each(_k_list.begin(), _k_list.end(), _do_reset);
}







