//============================================================================
// Project:       Bedrock - Foundational objects
// Filename:      ComponentLoader.cpp
// Description:   bedrock::Application implementation
// Created at:    Thu Jul 19 13:15:16 2001
// Modified at:   Mon Jul 30 19:32:10 2001
// 
// License:
// 
// The contents of this file are subject to the Jabber Open Source
// License Version 1.0 (the License). You may not copy or use this
// file, in either source code or executable form, except in compliance
// with the License. You may obtain a copy of the License at
// http://www.jabber.com/license/ or at http://www.opensource.org/.
// 
// Software distributed under the License is distributed on an AS IS
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and limitations
// under the License.
// 
// Copyrights:
// 
// Copyright (c) 1999-2001 Jabber.com
// 
// 
// $Id: ComponentLoader.cpp,v 1.1 2001/12/16 20:02:26 dr_pepper Exp $
//============================================================================
#include "jax.h"
#include <string>
using namespace jax;
using namespace std;

ComponentLoader::ComponentLoader(Component& ci) : 
    _ci(ci), _packet(0)
{}

ComponentLoader::~ComponentLoader() 
{}

int ComponentLoader::start(const char *hostname, unsigned int port, bool outgoing,
		     const char *component_name, const char *secret)
{
    bedrock::ThreadPool tpool(1);
    bedrock::net::SocketWatcher watcher(tpool, 2);
    _router = new RouterConnection<Packet, Packet>(watcher, *this, 
						   outgoing, 0);

    _name = component_name;
    _secret = secret;
    _outgoing = outgoing;
    _host = hostname;
    _port = port;

    if (outgoing)
    {
	bedrock::net::Address a(hostname, "", port);
	_router->connect(component_name, secret, a);
    }
    else
    {
	bedrock::net::Address a;
	a.addEntry(0, hostname, port);
	_router->connect(component_name, secret, a);
    }
    return bedrock::Application::start();
}

// EventListener
void ComponentLoader::onRouterConnected()
{
    string ns("jabber:config:");
    ns += _name;
    judo::Element *request = new judo::Element("xdb");
    request->putAttrib("type", "get");
    request->putAttrib("to", "config@-internal");
    request->putAttrib("from", _name);
    request->putAttrib("ns", ns.c_str());
    _router->deliver(new Packet(request));
}

void ComponentLoader::onRouterDisconnected()
{
    if (_packet == NULL)
    {
	bedrock::Application::exit(1, "Disconnected from router before receiving configuration");
    }

    // Give the router a moment to reset its listening port
    sleep(1);

    if (_packet->getElement()->empty())
    {
	_ci.init(0);
    }
    else
    {
	judo::Element *config = dynamic_cast<judo::Element*>(*_packet->getElement()->begin());
	// Add connection information to the config
	config->addElement("jax:component:name", _name);
	config->addElement("jax:component:secret", _secret);
	if (_outgoing)
	{
	    config->addElement("jax:component:outgoing", "true");
	}
	config->addElement("jax:component:host", _host);
	char buf[12];
	memset(buf, '\0', sizeof(buf));
	sprintf(buf, "%d", _port);
	config->addElement("jax:component:port", buf);
	_ci.init(config);
    }
    delete(_packet);
}

void ComponentLoader::onRouterError()
{
    bedrock::Application::stop(1, "Router connection in error -- exiting.");
}

void ComponentLoader::onRouterPacket(Packet* p)
{
    _packet = p;
    judo::Element *ep = p->getElement();
    if (ep->getAttrib("type") != "result")
    {
	bedrock::Application::exit(1, "Non-configuration packet received. Exiting.");
    }
    _router->disconnect();
}
