/**
 * 
 * Copyright (C) 2001 Alexandre. N. Safiullin <alex_@unis-ru.com> 
 * 
 * The contents of this file are subject to the GNU General Public License
 * Version 2 (the License). You may not copy or use this
 * file, in either source code or executable form, except in compliance
 * with the License.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

//     In addition, as a special exception, Alexandre N. Safiullin 
//     gives permission to link the code of this program with
//     the JECL (Jabber External Component Library, jabber.tigris.org)
//     (or with modified versions of JECL that use the
//     same license as JECL), and distribute linked combinations including
//     the two.  You must obey the GNU General Public License in all
//     respects for all of the code used other than JECL.  If you modify
//     this file, you may extend this exception to your version of the
//     file, but you are not obligated to do so.  If you do not wish to
//     do so, delete this exception statement from your version.


/* File: main.cpp -> a part of the "icqv7-t" project 
   Created: Tue Nov 27 23:27:17 2001
		by Alexandre N. Safiullin <alex_@unis-ru.com>
   Contains: implementation of main */

#include <SmsTrans.h>
#include <IcqTrans.h>

void usage()
{
    cerr << endl <<"Usage: " << endl
         << "icqv7-t -n <component id> -s <secret> -h <jabberd ip> -p <jabberd port> -m <sms id>" << endl
		 << "  component name - the name the other system knows this component by" << endl
		 << "  secret         - shared secret between the two systems" << endl
		 << "  jabberd ip     - the IP of a jabberd that this component will connect to" << endl
		 << "  jabberd port   - the port where a jabberd is listening for a connection" << endl
		 << "  sms id         - the name the other system knows sms-t by" 
		 << endl << endl;
    
    exit(64);
}

int main(int argc, char** argv)
{
	string       component_id = "icqv7ext";
    string       secret       = "mysecret";
    bool         outgoing     = true;

    string       jabberd_ip   = "127.0.0.1";
    unsigned int jabberd_port = 7000;
	string       sms_id        = "sms.localhost";

    int retval;

    while (1)
    {
        retval = getopt(argc, argv, "h:n:p:s:m:");

        // done
        if (retval == -1)
        {
            break;
        }

        switch (retval)
        {
        case 'h':
            jabberd_ip = optarg;
            break;
        case 'n':
            component_id = optarg;
            break;
        case 'p':
            jabberd_port = atoi(optarg);
            break;
        case 's':
            secret = optarg;
            break;
        case 'm':
            sms_id = optarg;
            break;
        case ':':
        case '?':
        default:
            usage();
            break;              
        }
    }
  
    cerr << "Starting component..."   <<endl;
    cerr << "\tComponent ID : " << component_id <<endl;
    cerr << "\tSecret       : " << secret << endl;
    cerr << "\tJabberd IP   : " << jabberd_ip << endl;
    cerr << "\tJabberd Port : " << jabberd_port << endl;

    IcqTrans icq_t(component_id, secret, jabberd_ip, jabberd_port, outgoing);
    SmsTrans sms_t(sms_id,secret,jabberd_ip,jabberd_port+1, outgoing, icq_t);

    bedrock::Application::start();
}



