/**
 * 
 * Copyright (C) 2001 Alexandre. N. Safiullin <alex_@unis-ru.com> 
 * 
 * The contents of this file are subject to the GNU General Public License
 * Version 2 (the License). You may not copy or use this
 * file, in either source code or executable form, except in compliance
 * with the License.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

//     In addition, as a special exception, Alexandre N. Safiullin 
//     gives permission to link the code of this program with
//     the JECL (Jabber External Component Library, jabber.tigris.org)
//     (or with modified versions of JECL that use the
//     same license as JECL), and distribute linked combinations including
//     the two.  You must obey the GNU General Public License in all
//     respects for all of the code used other than JECL.  If you modify
//     this file, you may extend this exception to your version of the
//     file, but you are not obligated to do so.  If you do not wish to
//     do so, delete this exception statement from your version.

/* File: utils.cpp -> a part of the "icqv7-t" project 
   Created: Fri Dec  7 20:04:57 2001
		by Alexandre N. Safiullin <alex_@unis-ru.com>
   Contains: implementation of utils */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <it_Client.h>

iconv_t utils::converter_2UTF8;
iconv_t utils::converter_unUTF8;

Status utils::show2status(const char* _show) {
	if(_show==NULL)return STATUS_ONLINE;
	string show=_show;
	if (show=="away")
        return STATUS_AWAY;
    if (show=="busy")
        return STATUS_OCCUPIED;
    if (show=="chat")
        return STATUS_FREEFORCHAT;
    if (show=="dnd")
        return STATUS_DND;
    if (show=="xa")
        return STATUS_NA;

    return STATUS_ONLINE;
}

string utils::ui2string(const unsigned int uin){
	char buffer[16];
	snprintf(buffer,16,"%lu",uin);
	return buffer;
}

unsigned int utils::string2ui(const char* s){
	if(s==NULL)return 0;
	return strtoul(s,NULL,10);
}

string utils::_2UTF8(const string& s){
	size_t in_length=s.size();

	if(in_length==0)return "";

	size_t out_length=in_length*2+1;

	char* in=strdup(s.c_str());
	char* out=new char[out_length+1];
	
	ICONV_CONST char* inptr=in;
	char* outptr=out;

	iconv(converter_2UTF8,&inptr,&in_length,&outptr,&out_length);
	if(in_length!=0){
		t<<"Troubles in 2UTF8"<<endl;
		return "";
	}
   
	*outptr=0;
	
	string result(out);
	delete out;
	free(in);

	return result;
			 
}

string utils::_unUTF8(const string& s){
	size_t in_length=s.size();

	if(in_length==0)return "";

	size_t out_length=in_length;

	char* in=strdup(s.c_str());
	char* out=new char[out_length+1];
	
	ICONV_CONST char* inptr=in;
	char *outptr=out;

	iconv(converter_unUTF8,&inptr,&in_length,&outptr,&out_length);
	if(in_length!=0){
		t<<"Troubles in unUTF8"<<endl;
		return "";
	}
   
	*outptr=0;
	
	string result(out);
	delete out;
	free(in);

	return result;
}

void it_Client::Pinger::start() {
	if(started)return;

	pthread_create(&_thread,NULL,it_Client::Pinger::beginProcessing,
				   this);
	started=true;
}

void it_Client::Pinger::stop() {
	if(!started)return;
	pthread_cancel(_thread);
	
	pthread_join(_thread, NULL);
	started=false;
	t<<"Pinging thread is finished"<<endl;
}

void* it_Client::Pinger::beginProcessing(void *arg) {
	it_Client::Pinger* pinger=(it_Client::Pinger*)arg;
	pinger->process();
	return NULL;
}

void it_Client::Pinger::process(void) {
	t<<"Entering ping loop"<<endl<<flush;
	unsigned int sec30=0;
	while (1) {
		sleep(5);
		sec30+=5;
		pthread_testcancel();
		if(sec30==30){
			t<<"Clearing out contacts cache"<<endl;
			session->contacts.clearoutPoll();
			sec30=0;
		}
		session->Poll();
	}
}
